'''
CF-1.0 checker base code

implementation of requirements and recommendations from
http://cf-pcmdi.llnl.gov/conformance/requirements-and-recommendations/1.0/ and
http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.0/cf-conventions.html

this code is used as a base for higher-version CF checks by overloading and extending the
respective test functions.

non-standard dependencies:
- NetCDF4-python
    - requires numpy
- lxml
- udunits (from unidata CDAT, http://www2-pcmdi.llnl.gov/cdat)
'''

import os
import string
import re
import datetime
import operator as op

from netCDF4 import Dataset as ncdf4
from lxml import etree
import numpy as np

from CFchecker import __version__
from CFchecker import iso_time
from CFchecker.log import MessageLog
from CFchecker.std_name_table import StandardNameTable

from errors import MessageLogOverflow, CFCheckerError, CellMethodsParseError
from utils import is_string_type, is_num_type, flatten, find_duplicates, map_axes, \
                            find_bounds, find_dim_vars, parse_cell_methods, is_valid_cf_name, \
                            find_climatology_vars, find_compress_vars, find_auxiliary_coord_vars, \
                            find_label_vars, find_grid_mapping_vars, map_dim_vars_to_types, \
                            units_known, units_convertible, is_ascending, is_monotonic

from CFchecker import _datadir

##################
##### tests ######
##################

class CF10Tests(object):
    cf_convention = 'CF-1.0'
    epsilon = pow(2,-53) # epsilon value to be used for float comparisons
    def __init__(self, name_table_file=os.path.join(_datadir, 'cf-standard-name-table.xml'),
                 not_implemented_info=False, max_entries=500, tests_to_run=None):
        '''
            name_table_file: full path of cf-standard-name-table.xml to be used
            not_implemented_info: log some messages that inform about checks that are not implemented (yet)
            max_entries: maximum number of entries in the log before checking will be aborted
            tests_to_run: list of test number strings or None if all tests should be run
        '''

        self.msg = MessageLog(log_not_implemented=not_implemented_info, max_entries=max_entries)
        self.all_tests = [name for name in dir(self) if name.startswith('_test_')]

        # check if all requested tests are available
        if tests_to_run is None: self.tests_to_run = self.all_tests
        else:
            self.tests_to_run = []
            for tstid in sorted(tests_to_run):
                test_name = '_test_%s' % tstid.replace('.', '_')
                if test_name not in self.all_tests:
                    raise CFCheckerError('testid "%s" is unknown' % (tstid,))
                else:
                    self.tests_to_run.append(test_name)

        # initizalize standard name table before use
        self.name_table = self._load_name_table(name_table_file)

    def __del__(self):
        '''destructor, closes ncfile'''
        try:
            self.ncfile.close()
        except:
            pass

    def _test_2_1(self):
        '''2.1 Filename
            Requirements:
            * Filename must have ".nc" suffix.
        '''
        if not self.filename.endswith('.nc'):
            self.msg.error('2.1', 'file does not have .nc extension')

    def _test_2_2(self):
        '''2.2 Data Types
            Requirements:
            * CF attributes that take string values must be 1D character arrays.
        '''
        # how to check for that?
        self.msg.not_implemented('2.2', 'not checking if CF string attributes are 1D char arrays')


    def _test_2_3(self):
        '''2.3 Naming Conventions
            Requirements:
            * Variable, dimension and attribute names must begin with a letter
              and be composed of letters, digits, and underscores.
              -- NOTE: assuming this refers to ASCII chars
            Recommendations:
            * No two variable names should be identical when case is ignored.
        '''
        attname_exceptions = ['_FillValue']
        # check global attribute names
        for att in self.ncfile.ncattrs():
            if not is_valid_cf_name(att):
                self.msg.error('2.3', 'global attribute name "%s" does not follow naming conventions' % att)
        # check dimension names
        for dim in self.ncfile.dimensions:
            if not is_valid_cf_name(dim):
                self.msg.error('2.3', 'dimension name "%s" does not follow naming conventions' % dim)
        # check variable names
        for vname in self.ncfile.variables:
            if not is_valid_cf_name(vname):
                self.msg.error('2.3', 'variable name "%s" does not follow naming conventions' % vname)
            # check variable's attribute names
            for attname in self.ncfile.variables[vname].ncattrs():
                # do not check special names from the exception list
                if attname in attname_exceptions: continue
                if not is_valid_cf_name(attname):
                    self.msg.error('2.3', 'attribute name "%s" of variable "%s" does not follow naming conventions'
                                                 % (attname, vname))
        # check for identical variable names when case is ignored
        vlist = self.ncfile.variables.keys()
        for idx1 in xrange(len(vlist)-1):
            for idx2 in xrange(idx1+1, len(vlist)):
                if vlist[idx1].upper() == vlist[idx2].upper():
                    # switch indices so that vlist[idx1] <= vlist[idx2] (more predictable log output)
                    if vlist[idx1] > vlist[idx2]:   idx1, idx2 = idx2, idx1
                    self.msg.warning('2.3', 'variable names "%s" and "%s" identical except for case'
                                                   % (vlist[idx1], vlist[idx2]))

    def _test_2_4(self):
        '''2.4 Dimensions
            Requirements:
            * The dimensions of a variable must all have different names.
            Recommendations:
            * If any or all of the dimensions of a variable have the interpretations
              (as given by their units or axis attribute) of time (T), height or depth (Z), latitude (Y),
               or longitude (X) then those dimensions should appear in the relative order T, then Z, then Y,
               then X in the CDL definition corresponding to the file.
            * In files that are meant to conform to the COARDS subset of CF, any dimensions of a variable other
              than space and time dimensions should be added "to the left" of the space and time dimensions as
              represented in CDL.
        '''
        # required order of coordinate dimensions in data variables:
        req_order = ['T','Z','Y','X']

        # first try to find axis attributes
        axis_map = map_axes(self.ncfile, reverse_map=True)
        # try to augment missing dimension variables in axis_map with guessed axis values from _deduce_coord_type
        for dimname in self.ncfile.dimensions:
            if not dimname in self.ncfile.variables:
                # no coordinate variable
                continue
            elif dimname in axis_map:
                # variable already in axis_map
                continue
            else:
                var = self.ncfile.variables[dimname]
                guess = self._deduce_coord_type(getattr(var, 'units', None), getattr(var, 'positive', None))
                if guess:
                    axis_map[dimname] = guess
        if not axis_map:    # no axes found
                self.msg.info('2.4', 'no axis information found in dimension variables, not checking dimension order')

        for vname in self.ncfile.variables:
            # check if variable depends on any dimensions at all (scalar variables have no dimensions)
            if len(self.ncfile.variables[vname].dimensions) == 0: continue
            duplicates = find_duplicates(self.ncfile.variables[vname].dimensions[:])
            if len(duplicates) > 0:
                # duplicates found
                for dim in duplicates:
                    self.msg.error('2.4', 'variable "%s" has multiple dependencies on dimension "%s"'
                                          % (vname, dim))
            else:
                # check for order of dimensions,
                # but skip those variables that have multiple dependencies anyway as they can't be correct here
                if axis_map:    # axes available found
                    # check if variable depends on any dimensions at all (scalar variables have no dimensions)
                    if len(self.ncfile.variables[vname].dimensions) > 0:
                        dim_axes = []
                        # generate list with order of axis names. all dimensions without axis names are listed as '*'
                        for dim in self.ncfile.variables[vname].dimensions:
                            dim_axes.append(axis_map.get(dim,'*'))
                        if set(dim_axes) == set('*'):  # no dimension with axis attribute found
                            # nothing to do, check next var
                            continue
                        # * at beginning should not occur
                        if dim_axes[0] == '*':
                            self.msg.warning('2.4', 'variable "%s": dimensions other than %s should be added to the left '
                                                    'according to COARDS, not to the right' % (vname, ','.join(req_order)))
                        # now remove all leading and trailing * for further testing
                        while dim_axes[0] == '*': del dim_axes[0]
                        # do the same for the end of the list
                        while dim_axes[-1] == '*':
                                del dim_axes[-1]
                        # now there should be only axis-dimensions left, check that
                        if '*' in dim_axes:
                            self.msg.warning('2.4', 'variable "%s": dimensions other than %s should be added to the left '
                                                    'according to COARDS, not in between' % (vname, ','.join(req_order)))
                            # now remove those *s too
                            while '*' in dim_axes: dim_axes.remove('*')
                        # convert remaining axis names to list of their index values according to req_order
                        dim_order = [req_order.index(axis) for axis in dim_axes]
                        # check if this list is ascending. in that case the order is correct
                        if not is_ascending(dim_order):
                            correct_order = [req_order[idx] for idx in sorted(set(dim_order))]
                            self.msg.warning('2.4', 'variable "%s": axis dimensions should be in following order: '
                                                    '%s and not: %s ' % (vname, ','.join(correct_order), ','.join(dim_axes)))

    def _test_2_5_1(self):
        '''
        2.5.1 Missing Data
        Requirements:
        * The valid_range attribute must not be present if the valid_min and/or valid_max
          attributes are present.
        * The _FillValue attribute must be the same type as its associated variable.
        * The missing_value attribute must be the same type as its associated variable.
        Recommendations:
        * The value of the _FillValue attribute should not be within a specified valid_range.
        * The missing_value attribute is deprecated and should no longer be used.
          However, if it's use is required for reasons of backwards compatibility with processing
          tools that only recognize the missing_value attribute, then we recommend that both
          missing_value and _FillValue be used, and they must have the same value.
          (starting with CF-1.5 missing_value is not deprecated any more, so we will not complain about it
          as it has never be a serious mistake to include it in the first place and it is even claimed that
          deprecation was a misunderstanding in the first place)
        ADDITIONS:
        * missing_value should also be checked against valid_range if there is no _FillValue

        dimension/coordinate variables are skipped in these tests as they should not have these attributes
        anyway. test 5 will check for presence in dimension variables.
        '''

        for vname in set(self.ncfile.variables) - set(self.dim_vars):
            var = self.ncfile.variables[vname]
            attrs = var.ncattrs()
            # check for valid_range and valid_min/valid_max attributes
            if 'valid_max' in attrs or 'valid_min' in attrs:
                # vaid_min and/or valid_max are present
                if 'valid_range' in attrs:
                    self.msg.error('2.5.1', '%s: valid_range attribute not allowed because ' \
                                            'valid_min and/or valid_max attributes are present.' % vname)
            # check _FillValue and missing_value types if available
            vartype = self.ncfile.variables[vname].dtype
            for attname in ( set(['_FillValue', 'missing_value']) & set(attrs) ):
                attval = getattr(var,attname)
                atttype = getattr(attval,'dtype', None)
                if not vartype == atttype:
                    self.msg.error('2.5.1', '%s: %s attribute and variable need to have the ' \
                                   'same data type (%s <-> %s)' % (vname, attname, atttype, vartype) )
            # check consistency of valid_range with _FillValue (or missing_value as fallback)
            fv_names = set(['_FillValue', 'missing_value']) & set(attrs)
            if 'valid_range' in attrs and len(fv_names) > 0:
                # if both attributes are present, check _FillValue, else whatever attribute is present
                fv_name = '_FillValue' if len(fv_names) == 2 else fv_names.pop()
                fv = getattr(var, fv_name)
                rmin = getattr(var, 'valid_range')[0]
                rmax = getattr(var, 'valid_range')[1]
                if fv >= rmin and fv <= rmax:
                    self.msg.warning('2.5.1', '%s: %s is within valid_range' % (vname, fv_name))
            # check for deprecated missing_value and for consistency of missing_value and _FillValue
            if set(['missing_value', '_FillValue']).issubset(set(attrs)):
                # will only check for equality of value
                # equality of type is handled by checking both att types against var type
                mv = getattr(var,'missing_value')
                fv = getattr(var,'_FillValue')
                if mv != fv:
                    self.msg.warning('2.5.1', '%s: attributes missing_value and _FillValue should ' \
                                     'have the same value (%s <-> %s)' % (vname, mv, fv) )

    def _test_2_6_1(self):
        '''
        2.6.1 Identification of Conventions
        Requirements:
        * Files that conform to the CF version 1.0 conventions must indicate this by setting the global
          Conventions attribute to the string value "CF-1.0".
        '''
        if 'Conventions' in self.ncfile.ncattrs():
            conventions = getattr(self.ncfile,'Conventions')
            if not conventions == self.cf_convention:
                self.msg.error('2.6.1', 'global Conventions attribute should be set to "%s", not "%s"' %
                                         (self.cf_convention, conventions) )
        else:
            self.msg.error('2.6.1', 'global Conventions attribute missing (should be set to "%s")' % self.cf_convention)

    def _test_2_6_2(self):
        '''
        2.6.2 Description of File Contents
        Requirements:
        * The title, history, institution, source, references, and comment attributes are all type string.
        Recommendations:
        * The title and history attributes are only defined as global attributes. If they are used as per variable
          attributes a CF compliant application should treat them exactly as it would treat any other unrecognized
          attribute.
        '''
        atts_to_check = ('title', 'history', 'institution', 'source', 'references', 'comment')
        # first check globally, then in variables
        for att in atts_to_check:
            # global:
            if att in self.ncfile.ncattrs():
                attval = getattr(self.ncfile, att)
                if not is_string_type(attval):
                    self.msg.error('2.6.2', 'global attribute "%s" has to be of type string' % att)
            # do not check title and history in variables (according to recommendations)
            if att in ('title', 'history'): continue
            # check in all variables
            for vname in self.ncfile.variables:
                if att in self.ncfile.variables[vname].ncattrs():
                    attval = getattr(self.ncfile.variables[vname], att)
                    if not is_string_type(attval):
                        self.msg.error('2.6.2', 'variable %s: attribute "%s" has to be of type string' % (vname, att))

    def _test_3(self):
        '''
        3 Description of the Data
        Recommendations:
        * All variables should use either the long_name or the standard_name attributes to describe their contents.
          Exceptions are boundary and climatology variables.
        * ADDITION: label variables are ignored as well
        '''
        for vname in set(self.ncfile.variables) - set(self.clim_vars) \
                 - set(self.label_vars)-set(flatten(self.bounds.values())) - set(flatten(self.grid_mappings.values())):
            if not 'long_name' in self.ncfile.variables[vname].ncattrs() and not 'standard_name' in self.ncfile.variables[vname].ncattrs():
                self.msg.warning('3', 'variable "%s" contains neither long_name nor standard_name attribute' % vname)

    def _test_3_1(self):
        '''
        3.1 Units
        Requirements:
        * The units attribute is required for all variables that represent dimensional quantities (except for boundary
          variables defined in section 7.1 and climatology variables defined in section 7.4).
        * The type of the units attribute is a string that must be recognizable by the udunits package. Exceptions are
          the units level, layer, and sigma_level.
        * The units of a variable that specifies a standard_name must be consistent with the units given in the standard
          name table. The units must also be consistent with a specified cell_methods attribute, if one is present.
        Recommendations:
        * The units level, layer, and sigma_level are deprecated.

        * ADDITION: string-type variables and compress index variables are skipped

        if the variable is recognized as a time dimension variable and the units could not be read,
        a hint to test 4.4 is made to get more detailed analysis
        this is unlikely to happen, though, as units are a vital part of identifying a dimension's type

        missing units will only be treated as an error if there is a standard_name that suggests units, else
        there will only be an info message about not having found units on the variable

        if latitude and longitude can be identified, units of those will be checked more closely in tests 4.1 and 4.2
        '''

        units_deprecated = ['level', 'layer', 'sigma_level']
        units_noudunits = ['level', 'layer', 'sigma_level'] # do not have to be recognized by udunits
        for vname in set(self.ncfile.variables) - set(self.clim_vars) - set(flatten(self.bounds.values())) \
                 - set(flatten(self.grid_mappings.values())) - set(self.compress_vars):
            var = self.ncfile.variables[vname]
            if var.dtype.char == 'S':
                # skip string-type variables, units do not make sense there anyway
                continue
            # need to get standard_name first because we need modifiers to know if we should throw missing units errors
            std_name = getattr(var, 'standard_name', None)
            if is_string_type(std_name):
                # extract modifier from standard_name
                std_name_modifier = ' '.join(std_name.split()[1:])
                # extract only standard name without modifiers (see below)
                std_name = std_name.split()[0]
            else:
                # not found or invalid type
                # the error message for that comes in 3.3, so it is not generated here
                std_name = None
                std_name_modifier = None
            if 'units' not in var.ncattrs():
                if std_name is None:
                    # no standard_Name given (-> also no modifier), so cannot know if units are required.
                    # only generate info message for user to check
                    self.msg.info('3.1', 'variable "%s" does not contain units attribute' % vname)
                elif self.name_table.exists(std_name):
                    # no units but valid standard_name found. if this standard_name requires units not compatible to "1",
                    # this is an error. invalid standard_names are being dealt with in test 3.3, so this is not
                    # reported here
                    std_units = self.name_table.get_units(std_name)
                    if not units_convertible(std_units, '1') and std_name_modifier not in ('status_flag',):
                        # std_name units are not compatible to dimensionless quantities and modifier is not status_flag
                        # which requires no units
                        self.msg.error('3.1', 'variable "%s" requires a units attribute compatible with given standard_name\'s units ("%s") ' \
                                              '(and modifier)' % (vname, std_units))
            elif not is_string_type(getattr(var,'units', None)):
                self.msg.error('3.1', 'units attribute in variable "%s" is not of type string' % vname)
            else:
                # units are there and they are of type string
                units = getattr(var,'units', None)
                if units in units_deprecated:
                    self.msg.warning('3.1', '%s: usage of unit "%s" is deprecated' % (vname, units) )
                elif not units in units_noudunits and not units_known(units):
                    # more verbose error message for time variables
                    if vname in self.time_dims:
                        msg_append = ', see test 4.4 for more details'
                    else:
                        msg_append = ''
                    self.msg.error('3.1', '%s: units "%s" not recognized%s' % (vname, units, msg_append) )
                else:
                    # units known, so check consistency with standard_name and cell_methods
                    # first try to parse cell_methods. checks are woven into the standard_name checks (see below)
                    if 'cell_methods' in var.ncattrs():
                        cell_methods = getattr(var,'cell_methods', None)
                        if not is_string_type(cell_methods):
                            cell_methods = None
                        try:
                            #try to parse cell_methods content
                            cell_methods = parse_cell_methods(cell_methods)
                            if len(cell_methods) == 0:
                                cell_methods = None
                            else:
                                cell_methods = cell_methods[-1] # last element of cell_methods array is the interesting one for unit checks
                        except CellMethodsParseError, err:
                                cell_methods = None
                    else:
                        cell_methods = None
                    if std_name is not None:
                        if self.name_table.exists(std_name):
                            # non-existing standard names are handled by the next test,
                            # so this test only checks existing names
                            # depending on modifiers we need to check a little differently
                            if std_name_modifier in ('', 'detection_minimum', 'standard_error'):
                                # no modifier or modifiers that require same units type
                                std_units = self.name_table.get_units(std_name)
                                if cell_methods is not None:
                                    methods = cell_methods[1]
                                else:
                                    methods = None
                                if methods == 'variance':
                                    # standard units have to be squared if last operation on data was variance
                                    std_units = "%s %s" % (std_units, std_units)
                                if not units_convertible(units, std_units, reftimeistime=True):
                                    self.msg.error('3.1', '%s: units are "%s" but should be "%s" according to standard name table ' \
                                                          '(and possibly modifier definition/cell methods)' % (vname, units, std_units))
                            elif std_name_modifier in ('number_of_observations'):
                                # this modifier requires unit '1'
                                # cell_methods should not be able to modify that
                                std_units = '1'
                                if not units_convertible(units, std_units, reftimeistime=True):
                                    self.msg.error('3.1', '%s: units are "%s" but should be "%s" according to modifier definition' % (vname, units, std_units) )
                            elif std_name_modifier in ('status_flag'):
                                # this should not have a units attribute
                                std_units = None
                                self.msg.error('3.1', '%s: status_flag modifier of standard_name does not use unit attribute' % vname)
                            else:
                                # don't know that modifier, test 3.3 should report that
                                std_units = None
                    else:
                        if cell_methods is not None:
                            # no standard_name found. that means cell_methods cannot really be checked
                            self.msg.info('3.1', '%s: cell_methods found but no standard_name, so not possible to check consistency of cell_methods with units' % vname)

    def _test_3_3(self):
        '''
        3.3 Standard Name
        Requirements:
        * The standard_name attribute takes a string value comprised of a standard name optionally followed by one or
          more blanks and a standard name modifier.
        * The legal values for the standard name are contained in the standard name table.
        * The legal values for the standard name modifier are contained in Appendix C, Standard Name Modifiers.
        '''
        valid_modifiers = ['', 'detection_minimum', 'number_of_observations', 'standard_error', 'status_flag']

        for vname, var in self.ncfile.variables.iteritems():
            if 'standard_name' in var.ncattrs():
                std_name = getattr(var, 'standard_name')
                if not is_string_type(std_name):
                    self.msg.error('3.3', '%s: standard_name attribute has to be of type string' % vname)
                else:
                    # we have to use re for splitting to be sure that we do not split on any whitespace but only
                    # on one or multiple spaces
                    std_name_splt = re.split(' +', std_name)
                    std_name_modifier = ' '.join(std_name_splt[1:]) # extract modifier from standard_name
                    std_name = std_name_splt[0] # extract only standard name without modifiers
                    if not self.name_table.exists(std_name):
                        self.msg.error('3.3', '%s: standard_name \'%s\' does not exist in standard name table' % (vname, std_name))
                    elif self.name_table.is_alias(std_name):
                        self.msg.warning('3.3', '%s: standard_name \'%s\' is an alias name. usage of the real name (%s) is recommended'
                                                 % (vname, std_name, self.name_table.lookup_alias(std_name)) )
                    if not std_name_modifier in valid_modifiers:
                        self.msg.error('3.3', '%s: standard_name modifier \'%s\' unknown' % (vname, std_name_modifier) )

    def _test_3_4(self):
        '''
        3.4. Ancillary Data
        [...]
        The attribute ancillary_variables is used to express these types of relationships.
        It is a string attribute whose value is a blank separated list of variable names. The nature of the relationship
        between variables associated via ancillary_variables must be determined by other attributes. The variables listed
        by the ancillary_variables attribute will often have the standard name of the variable which points to them including
        a modifier (Appendix C, Standard Name Modifiers) to indicate the relationship.
        '''
        for vname, var in self.ncfile.variables.iteritems():
            anc_vars = getattr(var, 'ancillary_variables', None)
            if anc_vars is not None:
                if not is_string_type(anc_vars):
                    anc_type = anc_vars.dtype if hasattr(anc_vars, 'dtype') else '<unknown>'
                    self.msg.error('3.4', '%s: ancillary_variables attribute should be of type string, not "%s"' % (vname, anc_type))
                else:
                    # try to parse variable list
                    anc_vars_list = anc_vars.split()
                    if len(anc_vars_list) == 0:
                        self.msg.warning('3.4', '%s: ancillary_variables attribute does have any usable content' % vname)
                    else:
                        # try to find the variables that are listed
                        for anc_varname in anc_vars_list:
                            if not anc_varname in self.ncfile.variables:
                                self.msg.error('3.4', '%s: referenced ancillary variable "%s" not found in file' % (vname, anc_varname))
                            else:
                                # variable available, now try to find out if standard_names are equal
                                # this is not mandatory, however and will only generate informative messages
                                anc_var = self.ncfile.variables[anc_varname]
                                std_name = getattr(var, 'standard_name', None)
                                anc_std_name = getattr(anc_var, 'standard_name', None)
                                # only check further if both variables have standard names
                                if std_name is not None and anc_std_name is not None:
                                    # only extract the standard name itself, without modifiers
                                    std_name = std_name.split()[0]
                                    anc_std_name = anc_std_name.split()[0]
                                    if std_name != anc_std_name:
                                        self.msg.info('3.4', '%s: standard_name of ancillary variable "%s" differs from standard_name in ' \
                                                             'this variable (%s <-> %s)' % (vname, anc_varname, anc_std_name, std_name))


    def _test_3_5(self):
        '''
        3.5 Flags
        Requirements:
        * The flag_values attribute must have the same type as the variable to which it is attached.
        ADDITIONS:
        * flag_meanings should be a string
        * flag_meanings should contain as many words as flag_values contains values
        * flag_values values should be unique within the set
        * only mentioned in table C1: standard_name attribute should have status_flag modifier
        '''
        for vname, var in self.ncfile.variables.iteritems():
            std_name = getattr(var, 'standard_name', None)
            if not is_string_type(std_name): std_name = ''
            std_name_split = std_name.split()
            has_status_flag = len(std_name_split) == 2 and std_name_split[1] == 'status_flag'
            fvals_found = 'flag_values' in var.ncattrs()
            fmean_found = 'flag_meanings' in var.ncattrs()
            # if flag_values or flag_meanings attributes are there without status_flag, warn about that
            if (fvals_found or fmean_found) and not has_status_flag:
                self.msg.warning('3.5', '%s: flag_* attributes found but status_flag modifier not set' % (vname,) )
            if fvals_found:
                fvals = getattr(var, 'flag_values')
                # string has no dtype and needs to be treated differently
                atttype = 'S' if is_string_type(fvals) else fvals.dtype.char
                vartype = var.dtype.char
                # if flag_values is a string, we convert it to a list of ints for checks
                if atttype == 'c':
                    fvals = [ord(c) for c in fvals]
                fvals_len = len(fvals)
                if not vartype == atttype:
                    self.msg.error('3.5', '%s: flag_values attribute and variable need to have the same data type (%s <-> %s)' % (vname, atttype, vartype) )
                if not fmean_found:
                    self.msg.error('3.5', '%s: flag_values attribute without flag_meanings attribute' % (vname,) )
                dupe_fvals = find_duplicates(fvals)
                if dupe_fvals:
                    self.msg.error('3.5', '%s: every flag_values value has to be unique within the set of all flag_values' % (vname,) )
            if fmean_found:
                fmean = getattr(var, 'flag_meanings')
                if not is_string_type(fmean):
                    self.msg.error('3.5', '%s: flag_meanings attribute should be a string' % vname)
                else:
                    if fvals_found:
                        # flag_meanings is always a string at this point
                        fmean = fmean.split()
                        fmean_len = len(fmean)
                        if not fvals_len == fmean_len:
                            self.msg.error('3.5', '%s: flag_values and flag_meanings attributes have different lengths (%s <-> %s)' % (vname, fvals_len, fmean_len) )
                    else:
                        self.msg.error('3.5', '%s: flag_meanings attribute without flag_values attribute' % vname)

    def _test_4(self):
        '''
        4 Coordinate Types
        Requirements:
        * The axis attribute may only be attached to a coordinate variable.
        * The only legal values of axis are X, Y, Z, and T (case insensitive).
        * The axis attribute must be consistent with the coordinate type deduced from units and positive.
        * A data variable must not have more than one coordinate variable with a particular value of the axis attribute.
        '''
        valid_axes = ('X', 'Y', 'Z', 'T')
        data_vars = list(set(self.ncfile.variables.keys()) - set(self.dim_vars))
        dim_to_axis = map_axes(self.ncfile, reverse_map=True)

        for vname, var in self.ncfile.variables.iteritems():
            # check all variables with axis attributes
            axis_value = getattr(var, 'axis', None)
            # do checks on axis attribute
            if axis_value is not None:
                if not is_string_type(axis_value):
                    self.msg.error('4', '%s: axis attribute has to be a string' % vname)
                elif vname not in self.dim_vars:
                    self.msg.error('4', '%s: axis attributes are only allowed for coordinate variables' % vname)
                else:
                    # axis value is string and variable is coordinate variable
                    if axis_value.upper() not in valid_axes:
                        self.msg.error('4', '%s: axis attribute contains invalid value: %s' % (vname, axis_value))
                    else:
                        # check if coordinate type can be deduced from units and positive
                        # if yes, then compare to actual axis attribute value
                        axis_deduced = self._deduce_coord_type(getattr(var, 'units', None), getattr(var, 'positive', None))
                        if  axis_deduced is not None:
                            if axis_deduced != axis_value:
                                self.msg.error('4', '%s: axis attribute inconsistent with type implied by units '
                                               'and/or positive attributes (%s <-> %s)' % (vname, axis_value, axis_deduced))
            # check data variables for dependencies on coordinate variables' axis attribute
            if vname in data_vars:
                var = self.ncfile.variables[vname]
                # build list of all dimension variables the data variable depends on
                dep_axes = []
                for dim in var.dimensions:
                    if dim in dim_to_axis:
                        dep_axes.append((dim_to_axis[dim], dim))
                # find duplicate axis identifiers in that list
                dupes = find_duplicates([item[0] for item in dep_axes])
                # report an error for every duplicate axis identifier found
                for axis_value in dupes:
                    # find all dimension variables with that axis
                    same_axis = [item[1] for item in dep_axes if item[0] == axis_value]
                    self.msg.error('4', '%s: variable depends on multiple coordinate variables with axis attriute "%s": %s' %
                                        (vname, axis_value, ','.join(same_axis)))

    def _common_test_4_1__4_2(self, dim_to_check, tst_id):
        '''
        common test function for tests 4.1 and 4.2 as they are pretty much the same except for the dimension
        that has to be checked

        dim_to_check: dimension to check, either "lat" or "lon"
        tst_id: string ID of the test run ('4.1' or '4.2')

        Constraints to be checked:
        * Variables representing lat|lon must always explicitly include the units attribute
        * The recommended unit of lat|lon is degrees_(north|east). Also acceptable are degree_(north|east),
          degree_(N|E), degrees_(N|E), degree(N|E), and degrees(N|E)
        * the lat|lon type may be indicated additionally by providing the axis attribute with the value Y|X
        '''
        if dim_to_check not in ('lat', 'lon'):
            raise CFCheckerError('unknown dimension to check: "%s"' % dim_to_check)

        dim_std_name = {'lat': 'latitude', 'lon': 'longitude'}[dim_to_check]
        dim_lng_name = {'lat': 'latitude', 'lon': 'longitude'}[dim_to_check]
        dim_axis_name = {'lat': 'Y', 'lon': 'X'}[dim_to_check]
        units_preferred = {'lat': 'degrees_north', 'lon': 'degrees_east'}[dim_to_check]
        units_accepted = { 'lat': (units_preferred, )
                                + ('degree_north', 'degree_N', 'degrees_N', 'degreeN', 'degreesN'),
                            'lon': (units_preferred, )
                                + ('degree_east', 'degree_E', 'degrees_E', 'degreeE', 'degreesE'),}[dim_to_check]

        log_func = {'std': self.msg.error, 'lng': self.msg.warning}

        for vname in self.dim_vars:
            var = self.ncfile.variables[vname]
            # need to find out if coordinate variable is supposed to be lat|lon, this can so far only
            # be done by finding the appropriate standard_name value or long_name value (which is more of a guess)
            std_name = getattr(var,'standard_name', None)
            long_name = getattr(var, 'long_name', None)
            if std_name == dim_std_name:
                # found standard_name, so this should make pretty clear what we are dealing with
                identified_by = 'std'
            elif is_string_type(long_name) and dim_lng_name in long_name:
                # found term in long_name, so there is a chance that this is a latitude|longitude dimension
                identified_by = 'lng'
            else:
                identified_by = None

            if identified_by is not None:
                problems_found = False
                # check the above constraints as we think this should be a latitude|longitude coordinate
                if 'units' not in var.ncattrs():
                    problems_found = True
                    log_func[identified_by](tst_id, '%s: %s coordinates must always explicitly include the units attribute'
                                                 % (vname, dim_std_name))
                else:
                    units = getattr(var,'units')
                    if units not in units_accepted:
                        # units are none of the accepted ones
                        problems_found = True
                        log_func[identified_by](tst_id, '%s: %s coordinates must have units of "%s" or compatible'
                                                     % (vname, dim_std_name, units_preferred))
                    elif units != units_preferred:
                        # units are among the accepted units, but other units would be recommended
                        problems_found = True
                        self.msg.warning(tst_id, '%s: use of %s coordinate units "%s" is recommended instead of "%s"'
                                                % (vname, dim_std_name, units_preferred, units))

                # if there is an axis value, it should be Y|X. non-string values will be reported by test 4 already
                axis = getattr(var, 'axis', None)
                if is_string_type(axis) and axis.upper() != dim_axis_name:
                    log_func[identified_by](tst_id, '%s: optional axis attribute of %s coordinate must be set to "%s", '\
                                            'not "%s"' % (vname, dim_std_name, dim_axis_name, axis))

                if identified_by == 'lng' and problems_found:
                    self.msg.info(tst_id, '%s: %s was identified using hints in long_name, so this guess might be wrong. ' \
                                         'if so, then the warnings related to this test and variable can probably be ignored'
                                         % (vname, dim_std_name))
    def _test_4_1(self):
        '''
        4.1 Latitude Coordinate
        Requirements:
        * Variables representing latitude must always explicitly include the units attribute
        * The recommended unit of latitude is degrees_north. Also acceptable are degree_north,
          degree_N, degrees_N, degreeN, and degreesN
        * the latitude type may be indicated additionally by providing the axis attribute with the value Y

        Note: the actual test is handled by a common test function for 4.1 and 4.2
        '''
        self._common_test_4_1__4_2('lat', '4.1')


    def _test_4_2(self):
        '''
        4.2 Longitude Coordinate
        Requirements:
        * Variables representing longitude must always explicitly include the units attribute
        * The recommended unit of longitude is degrees_east. Also acceptable are degree_east,
          degree_E, degrees_E, degreeE, and degreesE
        * the longitude type may be indicated additionally by providing the axis attribute with the value X

        Note: the actual test is handled by a common test function for 4.1 and 4.2
        '''
        self._common_test_4_1__4_2('lon', '4.2')

    def _test_4_3(self):
        '''
        4.3 Vertical (height or depth) Coordinate
        Requirements:

        * The only legal values for the positive attribute are up or down (case insensitive).
        '''
        for vname, var in self.ncfile.variables.iteritems():
            pos_val = getattr(var, 'positive', None)
            if pos_val is not None:
                if not (is_string_type(pos_val) and pos_val.lower() in ('up', 'down')):
                    self.msg.error('4.3', '%s: invalid positive value "%s", valid attributes are only "up" and "down"'
                                           % (vname, pos_val))

    def _test_4_3_2(self):
        '''
        4.3.2 Dimensionless Vertical Coordinates
        Requirements:
        * The formula_terms attribute is only allowed on a coordinate variable which has a standard_name
          listed in Appendix C.
        * The type of the formula_terms attribute is a string whose value is list of blank separated
          word pairs in the form term: var. The legal values term are contained in Appendix C for each
          valid standard_name. The values of var must be variables that exist in the file.
        ADDITIONS:
        * warn if standard_name is in list for known formulas but does not provide formula_terms
        '''
        known_formula_terms, formula_constraints = self._setup_formula_terms()
        formula_terms_re = re.compile('^(\w+: \w+ ?)+$')    # term: var, at least one occurence

        for vname, var in self.ncfile.variables.iteritems():
            formula_terms = getattr(var, 'formula_terms', None)
            std_name = getattr(var, 'standard_name', None)
            if formula_terms is not None:
                if not vname in self.ncfile.dimensions:
                    self.msg.error('4.3.2', '%s: formula_terms attribute only allowed on coordinate variables'
                                             % vname)
                    continue

                if std_name is None:
                    self.msg.error('4.3.2', '%s: cannot check correctness of formula_terms because standard_name '
                                            'is missing' % vname)
                    continue

                if not is_string_type(std_name):
                    # we will not handle this case here as another test should have taken care of this
                    continue

                if not std_name in known_formula_terms:
                    self.msg.error('4.3.2', '%s: standard_name "%s" does not allow formula_terms'
                                                % (vname, std_name))
                    continue

                if not formula_terms_re.match(formula_terms):
                    self.msg.error('4.3.2', '%s: formula_terms has invalid format' % vname)
                    continue

                # easy parsing due to passed regex check
                formula_terms = formula_terms.split()
                # remove trailing ":" from all term names
                terms = [term[:-1] for term in formula_terms[0::2]]
                variables = formula_terms[1::2]
                terms_found = False
                for known_terms in known_formula_terms[std_name]:
                    if known_terms.issubset(set(terms)):
                        terms_found = True
                        break
                if not terms_found:
                    all_terms = set()
                    for item in known_formula_terms[std_name]: all_terms = all_terms.union(item)
                    missing_terms = all_terms - set(terms)
                    self.msg.error('4.3.2', '%s: at least one required term could not be found in formula_terms. '
                                            'potentially missing terms: %s' % (vname, ', '.join(missing_terms)))

                # check if variables are contained in file
                missing_vars = set(variables) - set(self.ncfile.variables.keys())
                vars_found = True
                if missing_vars:
                    vars_found = False
                    self.msg.error('4.3.2', '%s: missing formula variable%s in file: %s'
                                            % (vname, 's' if len(missing_vars)>1 else '',
                                                ', '.join(list(missing_vars))))

                if terms_found and vars_found:
                    # all required terms and variables found and they are data variables, so do constraints check
                    terms_dict =  {}
                    for t,v in zip(terms, variables): terms_dict[t] = v
                    res = formula_constraints.get(std_name, lambda x, y, z: '')(self.ncfile, vname, terms_dict)
                    if res:
                        self.msg.error('4.3.2', '%s: detected violation of constraints for given standard_name: %s'
                                                % (vname, res))
            elif is_string_type(std_name) and std_name in known_formula_terms:
                # there is no formula_terms attribute in file but given standard_name should provide one
                self.msg.warning('4.3.2', '%s: standard_name "%s" suggests that there should be formula_terms, but none found'
                                          % (vname, std_name))

    def _test_4_4(self):
        '''
        4.4 Time Coordinate
        Requirements:
        * The time units of a time coordinate variable must contain a reference time.
        * The reference time of a time coordinate variable must be a legal time in the specified calendar.
        Recommendations:
        * The use of a reference time in the year 0 to indicate climatological time is deprecated.
          This restriction only applies to the real-world calendar as used by the udunits package.
        * Units of year and month and any equivalent units should be used with caution.

        this test might generate some errors that were already reported in a similar context in test 3.1
        the validity of calendar entries will not be checked at this point

        ADDITIONS:
        * if unit is year or month, then time indices should to be int values (or at least float without fractional value)
          because calendar years and months have no fixed duration and years and months defined in CF thus do not
          correspond with acutal calendar units
        '''
        re_time_units = re.compile('^(?P<unit>\w+) since (?P<time>.+)$')
        for vname in self.time_dims:
            var = self.ncfile.variables[vname]
            units = getattr(var, 'units', None)
            # do not report a missing units attribute again as this should have been
            # done by test 3.1
            if units is None:
                continue
            time_match = re_time_units.match(units)
            if not time_match:
                self.msg.error('4.4', '%s: units attribute of a time variable must have format '
                                      '"<unit> since <reference time>"' % vname)
                continue
            time_unit = time_match.groupdict()['unit']
            reference_time = time_match.groupdict()['time']
            # time unit must be known by udunits and be of type time
            if not units_known(time_unit):
                self.msg.error('4.4', '%s: units attribute contains an unknown unit: "%s"'
                                       % (vname, time_unit))
                continue
            if not units_convertible(time_unit, 'second', reftimeistime=True):
                # this is unlikely to be detected as units are a vital part of identifying a dimension's type
                self.msg.error('4.4', '%s: units attribute contains a unit that does not represent time: "%s"'
                                       % (vname, time_unit))
                continue
            if time_unit.lower() in ('year', 'years', 'month', 'months'):
                self.msg.warning('4.4', '%s: the unit "%s" should be used with caution, it is NOT a '
                                        'calendar month/year' % (vname, time_unit))
                # check if time indices are ints/fractionless
                idx_diffs = var[:] - var[:].astype(int)
                if not (idx_diffs == 0.0).all():
                    self.msg.warning('4.4', '%s: should not use fractional indices with unit "%s" as this notation '
                                            'is highly error-prone (it does not correspond with calendar units)' % (vname, time_unit))
            # check if reference time is for year 0
            if re.match('^0{1,4}(-0?1(-0?1(( |T)00?(:00?(:00?)?)?)?)?)?$', reference_time):
                self.msg.warning('4.4', '%s: The use of a reference time in the year 0 to indicate '
                                        'climatological time is deprecated' % vname)
            else:
                # try to parse reference time
                try:
                    reference_time = iso_time.parse_one_datetime(reference_time)
                except Exception, ex:
                    self.msg.error('4.4', '%s: unable to parse reference time of units attribute: %s'
                                           % (vname, str(ex)))
                    continue
        self.msg.not_implemented('4.4', 'not checking if specified times are valid in the given calendar')

    def _test_4_4_1(self):
        '''
        4.4.1 Calendar
        Requirements:
        * The attributes calendar, month_lengths, leap_year, and leap_month may only be attached to time
          coordinate variables.
        * The standardized values of the calendar attribute are gregorian, standard, proleptic_gregorian,
          noleap, 365_day, all_leap, 366_day, 360_day, julian, and none (case insensitive).
          If the calendar attribute is given a non-standard value, then the attribute month_lengths is required,
          along with leap_year and leap_month as appropriate.
        * The type of the month_lengths attribute must be an integer array of size 12.
        * The values of the leap_month attribute must be in the range 1-12.
        * The values of the leap_year and leap_month attributes are integer scalars.
        Recommendations:
        * The attribute leap_month should not appear unless the attribute leap_year is present.
        * The time coordinate should not cross the date 1582-10-15 when the default mixed Gregorian/Julian
          calendar is in use.
        '''

        calendar_atts = set(['calendar', 'month_lengths', 'leap_year', 'leap_month'])
        std_calendar_values = ['gregorian', 'standard']
        calendar_values = std_calendar_values + ['proleptic_gregorian', 'noleap', '365_day',
                                                 'all_leap', '366_day', '360_day', 'julian', 'none']
        re_std_calendar_values = re.compile('(^%s$)' % '|'.join(std_calendar_values), re.IGNORECASE)
        re_calendar_values = re.compile('(^%s$)' % '|'.join(calendar_values), re.IGNORECASE)

        for vname, var in self.ncfile.variables.iteritems():
            # find out if non-time variables contain any calendar-type attributes
            if vname not in self.time_dims:
                cal_atts_in_var = set(var.ncattrs()) & calendar_atts
                if cal_atts_in_var:
                    self.msg.error('4.4.1', '%s: the attributes %s may only be attached to time coordinate '
                                            'variables' % (vname, ','.join(cal_atts_in_var)))
            else:
                # check calendar_atts for correct use in time variables
                calendar = getattr(var, 'calendar', None)
                if (calendar is not None) and re_std_calendar_values.match(calendar):
                    # standard calendar found, need to check date crossing
                    # try to parse time information with iso_time module
                    units = getattr(var, 'units', None)
                    if units is None:
                        # no units found
                        self.msg.info('4.4.1', '%s: not checking date/time because units attribute is missing'
                                               % (vname))
                    else:
                        # units found
                        try:
                            # convert numpy array to python list as iso_time has problems with numpy arrays
                            parsed_time = iso_time.from_enum(units, var[:].tolist())
                        except Exception, ex:
                            # no message for parse errors related to "year is out of range" (climatological year 0)
                            if isinstance(ex, ValueError) and str(ex) == 'year is out of range':
                                pass
                            else:
                                append_str = ': "%s"' % (str(ex),) if str(ex) else ''
                                self.msg.error('4.4.1', '%s: failed to parse date/time%s' % (vname, append_str))
                        else:
                            # if there is a time range in that variable, check if it crosses the magical date line
                            time_range = parsed_time.get_range() if hasattr(parsed_time, 'get_range') else None
                            date_line = datetime.datetime(1582, 10, 15)
                            if time_range and time_range[0] < date_line and time_range[1] >= date_line:
                                self.msg.warning('4.4.1', '%s: The time coordinate should not cross the date '
                                                          '1582-10-15 when the default mixed Gregorian/Julian '
                                                          'calendar is in use' % vname)

                elif calendar and re_calendar_values.match(calendar):
                    # one of the other default calendars found
                    pass
                else:
                    # no or unknown calendar
                    month_lengths = getattr(var, 'month_lengths', None)
                    leap_year = getattr(var, 'leap_year', None)
                    leap_month = getattr(var, 'leap_month', None)

                    if month_lengths is None:
                        # no month_lengths present
                        self.msg.error('4.4.1', '%s: no or unknown calendar defined, so month_lengths attribute '
                                                'is required' % (vname,))
                    else:
                        # month_lengths present
                        ml_type = month_lengths.dtype if hasattr(month_lengths, 'dtype') else '<unknown>'
                        if ml_type not in ('int32', 'int64'):
                            self.msg.error('4.4.1', '%s: month_lengths attribute has to be of type int, not "%s"'
                                                    % (vname, ml_type))
                        elif len(month_lengths) != 12:
                            # type is correct, check len
                            self.msg.error('4.4.1', '%s: month_lengths attribute must contain exactly 12 values, '
                                                    'not %d' % (vname, len(month_lengths)))

                    # check leap_month type, len and values
                    if leap_month is not None:
                        # leap_month present
                        lm_type = getattr(leap_month, 'dtype', '<unknown>')
                        if not (lm_type in ('int32', 'int64') and (leap_month.shape in ( (),(1,)) )):
                            self.msg.error('4.4.1', '%s: leap_month attribute must be a scalar int value'
                                                    % (vname,))
                        elif leap_month not in xrange(1,13):
                            self.msg.error('4.4.1', '%s: leap_month value must be in range 1-12'
                                                     % (vname,))

                    if leap_year is not None:
                        # leap_year present
                        # check type and len
                        ly_type = getattr(leap_year, 'dtype', '<unknown>')
                        if not (ly_type in ('int32', 'int64') and (leap_year.shape in ( (), (1,) ) )):
                            self.msg.error('4.4.1', '%s: leap_year attribute must be a scalar int value'
                                                    % (vname,))
                        if leap_month is None:
                            # no leap_month present
                            self.msg.info('4.4.1', '%s: unknown or no calendar defined and leap_year present '
                                                   'but no leap_month attribute found, so february is assumed'
                                                   % (vname,))
                    else:
                        # no leap_year present
                        self.msg.info('4.4.1', '%s: unknown calendar defined and no leap_year attribute found, '
                                               'it is assumed that there are no leap years' % (vname,))
                        if leap_month is not None:
                            # leap_month present
                            self.msg.warning('4.4.1', '%s: unknown calendar defined and leap_month without leap_year '
                                                      'present, so leap_month will be ignored' % (vname,))

    def _test_5(self):
        '''
        5 Coordinate Systems
        Requirements:
        * All of a variable's dimensions that are latitude, longitude, vertical, or time dimensions must
          have corresponding coordinate variables. (no idea how to check that)
        * A coordinate variable must have values that are strictly monotonic (increasing or decreasing).
        * A coordinate variable must not have the _FillValue or missing_value attributes.
        * The type of the coordinates attribute is a string whose value is a blank separated list of
          variable names. All specified variable names must exist in the file.
        * The dimensions of each auxiliary coordinate must be a subset of the dimensions of the variable
          they are attached to. An exception is a label variable which will have a trailing dimension for the
          maximum string length.
        Recommendations:
        * The name of a multidimensional coordinate variable should not match the name of any of its dimensions.

        Notes:
        * coordinate attribute parsing is done from scratch to catch and report errors instead of ignoring them
        '''
        all_aux_coords = set()

        for vname, var in self.ncfile.variables.iteritems():
            # check all dim_vars for monotony, _FillValue/missing_value
            if vname in self.dim_vars:
                # check monotony
                vals = var[:]
                if not is_monotonic(vals):
                    self.msg.error('5', '%s: a coordinate variable must have values that are strictly monotonic'
                                        % vname)
                # check _FillValue and missing_value presence
                if '_FillValue' in var.ncattrs() or 'missing_value' in var.ncattrs():
                    self.msg.error('5', '%s: a coordinate variable must not have the _FillValue or missing_value '
                                        'attributes' % vname)
            # gather and check coordinate attributes
            coords_str = getattr(var, 'coordinates', None)
            if coords_str is not None:
                if not is_string_type(coords_str):
                    self.msg.error('5', '%s: coordinates attribute has to be of type string' % vname)
                else:
                    # extract coordinates attribute information and check
                    aux_coords = set(coords_str.split())
                    # check for variables that are referenced but do not exist in file
                    vars_no_exist = aux_coords - set(self.ncfile.variables)
                    # add all existing auxiliary variables to global list
                    all_aux_coords |= (aux_coords - vars_no_exist)
                    if len(vars_no_exist) > 0:
                        self.msg.error('5', '%s: following variables that are referenced in coordinates attribute '
                                            'do not exist: %s' % (vname, ','.join(vars_no_exist)))
                    # check if dimensions of every referenced coordinate variable are subset of data variable
                    # coordinate variables that could not be found are excluded from this test
                    var_dims = set(var.dimensions)
                    for aux_name in (aux_coords - vars_no_exist - set(self.label_vars)):
                        # label variables will be checked in 6.1, so they are excluded here
                        aux_dims = set(self.ncfile.variables[aux_name].dimensions)
                        if not aux_dims.issubset(var_dims):
                            self.msg.error('5', '%s: dimensions of auxiliary coordinate variable "%s" are not a '
                                                'subset of the data variable\'s dimensions' % (vname, aux_name))
        # check all auxiliary vars for matches between var name and dimension names (if var is multi-dimensional)
        for aux_name in all_aux_coords:
            var = self.ncfile.variables[aux_name]
            if var.ndim > 1:
                if aux_name in var.dimensions:
                    self.msg.warning('5', '%s: name of a multidimensional auxiliary coordinate variable should not '
                                          'match the name of any of its dimensions' % aux_name)

    def _test_5_6(self):
        '''
        5.6 Grid Mappings and Projections
        Requirements:
        * The type of the grid_mapping attribute is a string whose value is a single variable name.
        * The specified variable name (known as a grid mapping variable) must exist in the file.
        * The grid mapping variable must have the grid_mapping_name attribute.
          The legal values for the grid_mapping_name attribute are contained in Appendix F.
        Recommendations:
        * The grid mapping variable should have 0 dimensions.
        '''
        grid_mappings = self._setup_grid_mappings()

        grid_vars = set()

        # first find all grid mapping variables by searching grid_mapping attributes
        # do not use find_grid_mapping_vars for this as invalid attributes should be reported and not skipped
        for vname, var in self.ncfile.variables.iteritems():
            grid_mapping = getattr(var, 'grid_mapping', None)
            if grid_mapping is not None:
                # grid_mapping found:
                if not is_string_type(grid_mapping):
                    self.msg.error('5.6', '%s: grid_mapping attribute has to be of type string' % vname)
                elif not grid_mapping in self.ncfile.variables:
                    self.msg.error('5.6', '%s: value of grid_mapping attribute must be a single variable name '
                                          'present in this file' % vname)
                else:
                    # referenced variable was found in file
                    grid_vars.add(grid_mapping)

        # check validity of all found grid mapping variables
        for vname in grid_vars:
            var = self.ncfile.variables[vname]
            if var.ndim != 0:
                self.msg.warning('5.6', '%s: a grid mapping variable should have 0 dimensions' % vname)
            mapping_name = getattr(var, 'grid_mapping_name', None)
            if mapping_name is None:
                # no grid_mapping_name found
                self.msg.error('5.6', '%s: a grid mapping variable must have a grid_mapping_name attribute'
                                       % vname)
            elif not is_string_type(mapping_name):
                self.msg.error('5.6', '%s: grid_mapping_name attribute has to be of type string' % vname)
            elif mapping_name not in grid_mappings:
                self.msg.error('5.6', '%s: grid_mapping_name "%s" is unknown' % (vname, mapping_name))
            else:
                # mapping_name available and valid
                # check if only allowed attributes are used
                # grid_mapping_name has been found (or reported) already, so exclude that
                unknown_atts = set(var.ncattrs()) - set(['grid_mapping_name']) \
                                                        - set(grid_mappings[mapping_name])
                known_atts = set(var.ncattrs()) - unknown_atts - set(['grid_mapping_name'])
                if unknown_atts:
                    self.msg.error('5.6', '%s: the following attributes are not allowed for the given '
                                          'grid_mapping_name (%s): %s' % (vname, mapping_name, ','.join(unknown_atts)))
                if known_atts:
                    # run the available known attributes through the constraints check
                    # the check will write its own log entries
                    self._check_mapping_constraints(vname, known_atts)

    def _check_mapping_constraints(self, vname, attnames):
        '''
        outsourced grid_mapping attributes constraints checker for test 5.6 (see above)

        the function assumes that all attributes passed to it have already been checked for validity
        within the given grid_mapping_name

        var: variable reference the attributes belong to
        atts: attributes to check
        returns: nothing
        '''
        constraints = self._setup_grid_mapping_constraints()
        var = self.ncfile.variables[vname]

        for attname in attnames:
            if attname not in constraints:
                raise CFCheckerError('attribute %s not found in constraints dict even though it should be valid. '
                                     'this is a bug, please report it' % attname)
            # get attribute value, type should be either numpy array or string
            val = getattr(var, attname)

            # branch depending on type of attribute
            att_type = constraints[attname][0]
            if att_type == 'S':
                # string type, only check if attribute value is really a string
                if not is_string_type(val):
                    self.msg.error('5.6', '%s: grid_mapping attribute %s has to be of type string' % (vname, attname))
            elif att_type == 'N':
                # num type
                # get the limits
                l_bnd, u_bnd = constraints[attname][1]
                # get the size bounds
                l_size, u_size = constraints[attname][2]
                # need to distinguish scalar and vector cases for is_num_type
                # for scalar cases we test val directly, for vector cases we test val[0]
                # which should suffice as in numpy arrays the data type of every element is the same
                if (val.size == 1 and not is_num_type(val)) or \
                   (val.size > 1 and not is_num_type(val[0])):
                    self.msg.error('5.6', '%s: grid_mapping attribute %s has to be a numerical datatype'
                                           % (vname, attname))
                else:
                    # value has numerical datatype
                    # check size first, then values
                    if not (l_size <= val.size <= u_size):
                        self.msg.error('5.6', '%s: grid_mapping attribute %s\'s size is out of bounds'
                                              % (vname, attname))
                    elif not ((l_bnd <= val).all() and (val <= u_bnd).all()):
                        # this condition also catches the multidimensional cases
                        self.msg.error('5.6', '%s: grid_mapping attribute %s\'s value is out of bounds'
                                              % (vname, attname))
            else:
                raise CFCheckerError('invalid attribute type "%s" found in constraints dict, '
                                     'this is a bug, please report it' % att_type)

    def _test_6_1(self):
        '''
        6.1 Labels
        Requirements:
        * A variable of character type that is named by a coordinates attribute is a label variable.
          This variable must have two dimensions. The leading dimension (CDL order) must match one
          of those of the associated variable, and the trailing dimension is for the maximum string length.
        '''
        for vname, var in self.ncfile.variables.iteritems():
            coordinates = getattr(var, 'coordinates', None)
            if coordinates is not None and is_string_type(coordinates):
                # only check those coordinate variables that are actually there
                cnames = [cname for cname in coordinates.split() if cname in self.ncfile.variables]
                for cname in cnames:
                    coord_var = self.ncfile.variables[cname]
                    if coord_var.dtype.char == 'S':
                        # variable is string and referenced by coordinates attribute
                        # according to CF-1.0 this means it should be a label variable
                        if not coord_var.ndim == 2:
                            self.msg.error('6.1', '%s: label variables must have exactly 2 dimensions, not %d'
                                                   % (cname, coord_var.ndim))
                        elif not coord_var.dimensions[0] in var.dimensions:
                            # first dimension of coordinate variable has to be from referencing variable's dimensions
                            self.msg.error('6.1', '%s: leading dimension (CDL order) must match one of those of '
                                                  'the referencing variable' % cname)

    def _test_7_1(self):
        '''
        7.1 Cell Boundaries
        Requirements:
        * The type of the bounds attribute is a string whose value is a single variable name.
          The specified variable must exist in the file.
        * A boundary variable must have the same dimensions as its associated variable,
          plus have a trailing dimension (CDL order) for the maximum number of vertices in a cell.
        * A boundary variable must be a numeric data type.
        * If a boundary variable has units or standard_name attributes, they must agree with those
          of its associated variable.
        Recommendations:
        * The points specified by a coordinate or auxiliary coordinate variable should lie within,
          or on the boundary, of the cells specified by the associated boundary variable.
        * Boundary variables should not have the _FillValue or missing_value attributes.
        '''
        npdt = lambda t: [np.dtype(tt) for tt in t]
        bounds_vars = set()
        for vname, var in self.ncfile.variables.iteritems():
            bounds = getattr(var, 'bounds', None)
            if bounds is not None:
                # bounds attribute exists
                if not is_string_type(bounds):
                    self.msg.error('7.1', '%s: bounds attribute has to be of type string' % vname)
                elif bounds not in self.ncfile.variables:
                    self.msg.error('7.1', '%s: bounds variable "%s" not found in file' % (vname, bounds))
                else:
                    # bounds attribute found in file
                    bounds_var = self.ncfile.variables[bounds]
                    # check dimensions
                    var_dims = var.dimensions
                    bounds_dims = bounds_var.dimensions
                    if not bounds_dims[:-1] == var_dims:
                        self.msg.error('7.1', '%s: boundary variable %s must have the same dimensions as '
                                              'its associated variable, plus have a trailing dimension '
                                              'for the maximum number of vertices in a cell' % (vname, bounds))
                    # check type
                    elif not bounds_var.dtype in npdt( (np.float64, np.float32, np.int64, np.int32, np.int16, np.int8) ):
                        self.msg.error('7.1', '%s: boundary variable %s must be a numeric data type'
                                               % (vname, bounds))
                    else:
                        # check if data lies within bounds (only works for 2-sided cells on 1-dim and scalar variables so far)
                        num_sides = len(self.ncfile.dimensions[bounds_dims[-1]])
                        if num_sides == 2 and var.ndim in (0,1):
                            # embed variable and bounds data into 1-element list in case of scalar data to be able to do the same processing
                            var_data = var[:] if var.ndim >=1 else [var[:]]
                            bnds_data = bounds_var[:] if var.ndim >= 1 else [bounds_var[:]]
                            for idx, (bd, vd) in enumerate(zip(bnds_data, var_data)):
                                if not ((bd[0] <= vd <= bd[1]) or (bd[1] <= vd <= bd[0])):
                                    self.msg.warning('7.1', '%s: data is outside of boundaries for at least one cell '
                                                            '(index %d)' % (vname, idx))
                                    break
                        else:
                            self.msg.info('7.1', '%s: not checking if data lies within bounds of %d-sided '
                                                            'cells of %d-dimensional variables'
                                                            % (vname, num_sides, var.ndim))

                    # check standard_name and units
                    # only check if attributes are present in both files
                    for att in (set(var.ncattrs()) & set(bounds_var.ncattrs())
                                & set(['standard_name', 'units'])):
                        if not getattr(var, att) == getattr(bounds_var, att):
                            self.msg.error('7.1', '%s: value of %s attribute has to be equal in data and '
                                                  'bounds variable' % (vname, att))
                    # remember bounds name for other checks
                    bounds_vars.add(bounds)

        # only check every bounds variable
        for bname in bounds_vars:
            bounds_var = self.ncfile.variables[bname]
            # check for _FillValue and missing_value
            bad_atts = set(bounds_var.ncattrs()) & set(['_FillValue', 'missing_value'])
            if bad_atts:
                self.msg.warning('7.1', '%s: boundary variables should not have the _FillValue or missing_value '
                                        'attributes' % bname)

    def _test_7_2(self):
        '''
        7.2 Cell Measures
        Requirements:
        * The type of the cell_measures attribute is a string whose value is list of blank separated word
          pairs in the form measure: var. The valid values for measure are area or volume.
          The var token specifies a variable that must exist in the file. The dimensions of the variable
          specified by var must be the same as, or be a subset of, the dimensions of the variable to which
          they are related.
        * A measure variable must have units that are consistent with the measure type, i.e., square meters
          for area measures and cubic meters for volume measures.
        '''
        for vname, var in self.ncfile.variables.iteritems():
            cell_measures = getattr(var, 'cell_measures', None)
            if cell_measures is not None:
                # cell_measures found
                if not is_string_type(cell_measures):
                    self.msg.error('7.2', '%s: cell_measures attribute has to be of type string' % vname)
                else:
                    # parse contents
                    if not re.match('^((volume|area): \S+ ?)+$', cell_measures):
                        self.msg.error('7.2', '%s: cell_measures format has to be "measure: var [measure: var ...]"'
                                               % vname)
                    else:
                        # format is valid
                        parts = cell_measures.split()
                        req_units = {'volume': 'm3', 'area': 'm2'}
                        for measure, meas_vname in zip(parts[0::2], parts[1::2]):
                            #remove : in measure
                            measure = measure[:-1]
                            if not meas_vname in self.ncfile.variables:
                                self.msg.error('7.2', '%s: measure variable "%s" not found in file'
                                                      % (vname, meas_vname))
                            elif not set(self.ncfile.variables[meas_vname].dimensions).issubset(set(var.dimensions)):
                                self.msg.error('7.2', '%s: dimensions of measure variable have to be subset of '
                                                      'related variable\'s dimensions' % vname)
                            else:
                                meas_var = self.ncfile.variables[meas_vname]
                                units = getattr(meas_var, 'units', None)
                                if units is None or not is_string_type(units):
                                    # silently drop missing, non-string and unknown units,
                                    # those errors should be reported by test 3.1
                                    pass
                                elif not units_convertible(units, req_units[measure], reftimeistime=True):
                                    # got known units attribute, but units don't fit measure type
                                    self.msg.error('7.2', '%s: measure variable\'s units (%s) are not consistent with '
                                                          'given measure type (%s)' % (vname, units, measure))


    def _test_7_3(self):
        '''
        7.3 Cell Methods
        Requirements:
        * The type of the cell_methods attribute is a string whose value is one or more blank separated word
          lists, each with the form dim1: [dim2: [dim3: ...] ] method [(comment)] where brackets indicate
          optional words. The valid values for dim1 [dim2 [dim3 ...] ] are dimension names of the associated
          variable or valid standard names.
          The valid values of method are contained in Appendix D. When the method refers to a climatological
          time axis, the values in Appendix D may have the suffixes "within days", "over days", "within years",
          or "over years" or appended.
        * A given dimension name may only occur once in a cell_methods string. An exception is a climatological
          time dimension.
        * The comment, if present, must take the form
          ([interval: value unit [interval: ...] comment:] remainder)
          The remainder text is not standardized. If no interval clauses are present, the entire comment is
          therefore not standardized. There may be zero interval clauses, one interval clause,
          or exactly as many interval clauses as there are dims to which the method applies.
          The value must be a valid number and the unit a string that is recognizable by the udunits package.

        a CF-1.4 cell_methods parser will be used to parse the string even for CF-1.0.
        the only change that was not a clarification that matters here is the addition of the where modifier
        this will be explicitly checked here and is considered forbidden in CF < 1.4
        '''
        re_clim_modifier = re.compile('(over|within) (years|days)$')
        for vname, var in self.ncfile.variables.iteritems():
            cell_methods = getattr(var, 'cell_methods', None)
            if cell_methods is not None:
                #cell_methods found
                if not is_string_type(cell_methods):
                    self.msg.error('7.3', '%s: cell_methods has to be of type string' % vname)
                else:
                    try:
                        cell_methods = parse_cell_methods(cell_methods)
                    except CellMethodsParseError, err:
                        self.msg.error('7.3', '%s: cell_methods could not be parsed: %s' % (vname, str(err)))
                    else:
                        if len(cell_methods) == 0:
                            self.msg.error('7.3', '%s: cell_methods attribute given but no content found' % vname)
                        else:
                            # extract all dimensions from parsed cell_methods
                            all_dims = flatten([elem[0] for elem in cell_methods])
                            # find climatological time dimensions
                            clim_dims = set(flatten([elem[0] for elem in cell_methods if re_clim_modifier.match(elem[2])]))
                            # find duplicate dimension names and complain if it is not a climatological time dimension
                            dupe_dims = find_duplicates(all_dims)
                            for dupe_dim in dupe_dims:
                                if not dupe_dim in clim_dims:
                                    self.msg.error('7.3', '%s: dimension name "%s" is only allowed to occur once in '
                                                          'cell_methods string' % (vname, dupe_dim))
                            # check if all specified dimensions are associated with the variable of if
                            # specified dimension is a known standard_name
                            for dim in set(all_dims):
                                if not (dim in var.dimensions or self.name_table.exists(dim)):
                                    self.msg.error('7.3', '%s: cell_methods dimension "%s" has to be a dimension '
                                                          'of the associated variable or a valid standard name'
                                                          % (vname, dim))
                            # check if where keyword is appearing in modifier, that is not allowed in CF < 1.4
                            modifiers = (elem[2] for elem in cell_methods)
                            for modifier in modifiers:
                                if modifier.startswith('where'):
                                    self.msg.error('7.3', '%s: cell_methods modifier "where" not allowed in CF < 1.4'
                                                          % (vname,))
                            # call outsourced comment check
                            self._check_cell_methods_comments(vname, cell_methods)

    def _check_cell_methods_comments(self, vname, cell_methods):
        '''
        outsourced cell_methods comments check for test 7.3 (see above)

        vname: variable name the cell_methods belong to
        cell_methods: cell_methods structure as returned by parse_cell_methods
        returns: nothing
        '''
        re_intervals_and_comment = re.compile('^(interval: \S+ \w+ ?)+(comment: ?(.*))?$')
        re_num = re.compile('^(\d+(\.\d*)?)$')
        # get comments, exclude empty comments
        comments = [elem[3] for elem in cell_methods if len(elem[3]) > 0]
        for comment in comments:
            if re_intervals_and_comment.match(comment):
                # there is valid interval syntax in the string, possibly followed by free-form comment
                # so check the units of all intervals
                parts = comment.split()
                while len(parts) >= 3 and parts[0] == 'interval:':
                    num = parts[1]
                    if not re_num.match(num):
                        self.msg.error('7.3', '%s: in cell_methods comment: interval length "%s" not '
                                              'a valid number' % (vname, num))
                    units = parts[2]
                    if not units_known(units):
                        self.msg.error('7.3', '%s: in cell_methods comment: interval unit "%s" not '
                                              'recognized' % (vname, units))
                    parts = parts[3:]   # remove first 3 elements from parts
                # all interval elements should have been checked now
                if len(parts) > 0 and not parts[0] == 'comment:':
                    # this is not supposed to happen, but we never know...
                    self.msg.error('7.3', '%s: cell_methods comments could not be parsed: expected "comment:" '
                                          'but got "%s"' % (vname, parts[0]))
            elif comment.startswith('interval:'):
                # regex did not match but there is an interval clause, so that must be malformed
                self.msg.error('7.3', '%s: cell_methods comments could not be parsed: malformed interval clause '
                                          'found' % (vname))

    def _test_7_4(self):
        '''
        7.4 Climatological Statistics
        Requirements:
        * The climatology attribute may only be attached to a time coordinate variable.
        * The type of the climatology attribute is a string whose value is a single variable name.
          The specified variable must exist in the file.
        * A climatology variable must have the same dimension as its associated time coordinate variable,
          and have a trailing dimension (CDL order) of size 2.
        * A climatology variable must be a numeric data type.
        * If a climatology variable has units, standard_name, or calendar attributes,
          they must agree with those of its associated variable.
        * A climatology variable must not have _FillValue or missing_value attributes.
        '''
        clim_vars = set()
        for vname, var in self.ncfile.variables.iteritems():
            climatology = getattr(var, 'climatology', None)
            if climatology is not None:
                # climatology attribute found
                if not is_string_type(climatology):
                    self.msg.error('7.4', '%s: climatology attribute has to be of type string' % vname)
                else:
                    # climatology attribute is a string
                    units = getattr(var, 'units', None)
                    std_name = getattr(var, 'standard_name', None)
                    if (vname not in self.dim_vars) or \
                        (units is not None and is_string_type(units) and units_known(units) and \
                        not units_convertible(units, 'seconds', reftimeistime=True)) or \
                        (std_name is not None and is_string_type(std_name) and self.name_table.exists(std_name) and \
                        not units_convertible(self.name_table.get_units(std_name), 'seconds', reftimeistime=True)):
                        # variable is not recognized dimension/coordinate variable or
                        # units are a known unit type or standard_name is valid
                        # (all other errors about units will be treated in 3.x tests)
                        # so we check if units represent a time, else the climatology attribute is not allowed
                        # this won't catch all violations as there does not seem to be a safe way to identify
                        # a coordinate variable in the first place
                        self.msg.error('7.4', '%s: climatology attribute may only be attached to a time '
                                              'coordinate variable' % vname)
                    if climatology not in self.ncfile.variables:
                        self.msg.error('7.4', '%s: referenced climatology variable "%s" not found in file'
                                              % (vname, climatology))
                    else:
                        # referenced climatology variable found in file
                        clim_vars.add(climatology)
                        var_dims = var.dimensions
                        clim_var = self.ncfile.variables[climatology]
                        clim_var_dims = clim_var.dimensions
                        # check dimensions
                        if not (var_dims == clim_var_dims[:-1] and len(self.ncfile.dimensions[clim_var_dims[-1]]) == 2):
                            self.msg.error('7.4', '%s: climatology variable %s must have the same dimension as its '
                                                  'associated time coordinate variable, and have a trailing dimension '
                                                  '(CDL order) of size 2.' % (vname, climatology))
                        for attname in (set(var.ncattrs()) & set(clim_var.ncattrs()) \
                                        & set(['units', 'standard_name', 'calendar'])):
                            if not (getattr(var, attname) == getattr(clim_var, attname)):
                                self.msg.error('7.4', '%s: %s attribute values of coordinate and climatology variables '
                                                      'must agree with each other' % (vname, attname))

        for clim_vname in clim_vars:
            clim_var = self.ncfile.variables[clim_vname]
            if (set(clim_var.ncattrs()) & set(['_FillValue', 'missing_value'])):
                # _FillValue and/or missing_value found
                self.msg.error('7.4', '%s: a climatology variable must not have _FillValue or missing_value attributes'
                                      % clim_vname)
            # check datatype (string is the only non-numeric data type at this time)
            if clim_var.dtype.char == 'S':
                self.msg.error('7.4', '%s: climatology variable must be a numeric data type' % clim_vname)

    # TODO: do not check any kind of variable value if data is packed or compressed
    # or even better: try to unpack/uncompress before checking (might be difficult if CF-convention is not followed)
    # for now: just print an info if packaged data was found that might influence the outcome of other tests
    # as long as there is no decompress/unpack support
    def _test_8_1(self):
        '''
        8.1 Packed Data
        Requirements:
        * The scale_factor and add_offset attributes must be the same numeric data type.
        * If scale_factor and add_offset are a different type than the variable, then they must be either
          type float or type double.
        * If scale_factor and add_offset are a different type than the variable, then the variable must be
          type byte, short or int.
        Recommendations:
        * If scale_factor and add_offset are type float, the variable should not be of type int.
        '''
        packed_vars = []
        for vname, var in self.ncfile.variables.iteritems():
            packed_atts = set(var.ncattrs()) & set(['scale_factor',  'add_offset'])
            if packed_atts:
                # attributes found that suggest packed data
                packed_vars.append(vname)
                atts_strtype = [is_string_type(getattr(var, att)) for att in packed_atts]
                if True in atts_strtype:
                        self.msg.error('8.1','%s: scale_factor and add_offset attributes must be of a numeric data type'
                                              % vname)
                else:
                    # should be numeric types here
                    scale_factor = getattr(var, 'scale_factor', None)
                    add_offset = getattr(var, 'add_offset', None)
                    sf_type = scale_factor.dtype if scale_factor is not None else None
                    ao_type = add_offset.dtype if add_offset is not None else None
                    var_type = var.dtype
                    if len(packed_atts)>1:
                        # both attributes exist
                        if not sf_type == ao_type:
                            self.msg.error('8.1','%s: scale_factor and add_offset attributes must be of same numeric data '
                                                 'type' % vname)
                    if (sf_type != var_type ) or \
                       (ao_type != var_type):
                        # attribute data types different from variable type
                        if var_type.char not in ('b', 'B', 'h', 'H', 'i', 'q'):
                            # int8, int16, int32 and their unsigned versions
                            self.msg.error('8.1', '%s: scale_factor and/or add_offset data types are different from '
                                                  'variable data type, so variable data type must be byte, short or int'
                                                    % vname)
                        if (hasattr(sf_type, 'char') and sf_type.char not in ('f', 'd')) or \
                           (hasattr(ao_type, 'char') and ao_type.char not in ('f', 'd')):
                            self.msg.error('8.1', '%s: scale_factor and/or add_offset data types are different from '
                                                  'variable data type, then attribute data type must be either float '
                                                  'or double' % vname)
                    if ((hasattr(sf_type, 'char') and sf_type.char in ('f', )) or \
                        (hasattr(ao_type, 'char') and ao_type.char in ('f',))) and var_type.char in ('i', 'q'):
                        self.msg.warning('8.1', '%s: if scale_factor and/or add_offset are type float, the variable should not '
                                                'be of type int' % vname)
        if packed_vars:
            self.msg.info('8.1','following variables seem to contain packed data: %s. Messages from tests 4.4.1 and 7.1 '
                                'concerning these variables might therefore be incorrect' % ', '.join(sorted(packed_vars)))
    def _test_8_2(self):
        '''
        8.2 Compression by Gathering
        Requirements:
        * The compress attribute may only be attached to a coordinate variable with an integer data type.
        * The type of the compress attribute is a string whose value is a blank separated list of dimension names.
        The specified dimensions must exist in the file.
        * The values of the associated coordinate variable must be in the range starting with 0 and going up to the
        product of the compressed dimension sizes minus 1 (CDL index conventions).
        '''
        for vname, var in self.ncfile.variables.iteritems():
            compress = getattr(var, 'compress', None)
            if compress is not None:
                # compress attribute found
                if not is_string_type(compress):
                    self.msg.error('8.2', '%s: compress attribute must be of type string' % vname)
                elif vname not in self.dim_vars:
                    self.msg.error('8.2', '%s: compress attribute may only be attached to coordinate variables' % vname)
                elif var.dtype is not np.dtype(np.int32):
                    self.msg.error('8.2', '%s: compress attribute may only be attached to coordinate variables '
                                          'with integer data type' % vname)
                else:
                    # check attribute contents
                    compressed_dims = compress.split()
                    not_found_dims = set(compressed_dims) - set(self.ncfile.dimensions)
                    if len(compressed_dims) < 1:
                        self.msg.error('8.2', '%s: at least one dimension name is required in compress attribute' % vname)
                    elif not_found_dims:
                        self.msg.error('8.2', '%s: following dimension names referenced in compress attribute were not '
                                              'found in input file: %s' % (vname, ', '.join(not_found_dims)))
                    else:
                        # all dims found in file and we have at least one dimension
                        dim_product = reduce(op.mul, (len(self.ncfile.dimensions[dim]) for dim in compressed_dims))
                        vals = var[:]
                        if vals.min() < 0 or vals.max() > dim_product-1:
                            self.msg.error('8.2', '%s: values of variable must be in range [0;%d]' % (vname, dim_product-1))

    def _load_name_table(self, name_table_file):
        name_table = StandardNameTable(name_table_file)
        return name_table

    @classmethod
    def _deduce_coord_type(cls, units, positive=None):
        '''
        try to deduce coordinate variable type (X,Y,Z,T) from units and positive attributes
        '''
        # additional (deprecated) units for Z axis
        z_units = ('level','layer','sigma_level')

        # if units are unknown, we won't guess anything
        if not units or not is_string_type(units) or not (units_known(units) or units in z_units):
            return None

        # guesses for Z: units are form of pressure, positive is either up or down,
        #                ('level','layer','sigma_level') units
        if is_string_type(positive) and positive.lower() in ('up', 'down'):
            # only deduce 'Z' if units without positive do not conclude something else
            if cls._deduce_coord_type(units, None) in (None, 'Z'):
                return 'Z'
            else:
                # not conclusive as units and positive indicate different interpretations
                return None
        if units in z_units or units_convertible(units, 'hPa', reftimeistime=True):
            return 'Z'

        # after this point: positive should not appear and will invalidate the deduction
        if positive:
            return None

        # guesses for X: units are some form of degrees east, no positive
        x_units = ('degrees_east', 'degree_east', 'degrees_E', 'degree_E', 'degreesE', 'degreeE')
        if units in x_units:
            return 'X'

        # guesses for Y: units are some form of degrees north, no positive
        y_units = ('degrees_north', 'degree_north', 'degrees_N', 'degree_N', 'degreesN', 'degreeN')
        if units in y_units:
            return 'Y'

        # guesses for T: units are form of relative time, no positive
        if units_convertible(units, 'days since 1970-01-01', reftimeistime=False):
            return 'T'

        # no type could be deduced
        return None

    @classmethod
    def _setup_formula_terms(cls):
        '''
        generate dict of standard names with their associated formula_terms requirements
        values of the dict are tuples that contain sets of required terms. one complete set has to match
        to satisfy requirements.

        also generate a second dict of standard names with their associated constraints.
        standard names that do not impose any constraints are left away.
        the check is implemented by a function that is called by the check procedure. arguments should be:
        - ncfile handle
        - variable name being checked
        - dict with term->variable mapping
        the check function should only be called after existence of required formula_terms has been established!
        '''
        formula_terms = {'atmosphere_ln_pressure_coordinate': (set(['p0', 'lev']),),
                    'atmosphere_sigma_coordinate': (set(['ptop', 'sigma', 'ps']),),
                    'atmosphere_hybrid_sigma_pressure_coordinate': (set(['a', 'b', 'ps', 'p0']), set(['ap', 'b', 'ps'])),
                    'atmosphere_hybrid_height_coordinate': (set(['a' , 'b', 'orog']),),
                    'atmosphere_sleve_coordinate': (set(['a', 'b1', 'b2', 'ztop', 'zsurf1', 'zsurf2']),),
                    'ocean_sigma_coordinate': (set(['sigma', 'eta', 'depth']),),
                    'ocean_s_coordinate': (set(['s', 'eta', 'depth', 'a', 'b', 'depth_c']),),
                    'ocean_sigma_z_coordinate': (set(['s', 'eta', 'depth', 'depth_c', 'zlev', 'nsigma']),),
                    'ocean_double_sigma_coordinate': (set(['sigma', 'depth', 'z1', 'z2', 'a', 'href', 'k_c']),),
                    }

        # embedded constraints check functions
        def hbsigma_constraints(ncfile, vname, terms_mapping):
            '''
            check for atmosphere_hybrid_sigma_pressure_coordinate
            constraints are:
                - a,b depend on the same one dimension as the lev variable;
                  that lev is a dimension variable is already checked for and thus a and b's dimensions
                  have to have the same name as the lev variable effectively;
                  also implies that those 3 variables are 1-dimensional as this is goes for all dimension variables
                - p0 must be scalar or 1-dimensional with len 1
                - ps must be 2-3-dimensional
                - a + b == lev
            '''
            # epsilon for float compares
            eps = 1e-5

            # p0 has to be checked before it's possibly used in calculations,
            # if p0 is not present (valid for ap notation), then we cannot check everything here
            p0_name = terms_mapping.get('p0')

            if p0_name is not None:
                p0 = ncfile.variables[p0_name][:]
                if not (p0.ndim == 0 or (p0.ndim, len(p0)) == (1,1)):
                    return '%s(term p0) must be scalar or 1-dimensional with length 1' % p0_name

            # get the other needed variable values
            # there could be the term a or ap in the file, so we have to consider both possibilities
            if 'a' in terms_mapping:
                a_term = 'a'
                a_name = terms_mapping[a_term]
                a = ncfile.variables[a_name][:]
            elif p0_name is not None:
                # if not a, then ap has to be present. use ap and p0 to calculate a
                a_term = 'ap'
                a = ncfile.variables[terms_mapping['ap']][:] / p0
                a_name = '%s/%s' % (terms_mapping['ap'], terms_mapping['p0'])
            else:
                # could not get a (because p0 not available and ap given)
                a_term = None

            if a_term is not None:
                a_dims = ncfile.variables[terms_mapping[a_term]].dimensions

            b_name = terms_mapping['b']
            b = ncfile.variables[b_name][:]
            b_dims = ncfile.variables[b_name].dimensions
            ps_name = terms_mapping['ps']
            ps = ncfile.variables[ps_name][:]

            lev = ncfile.variables[vname][:]
            lev_dims = ncfile.variables[vname].dimensions

            # check constraints
            # same single dimension name for a,b that is the same as the lev variable name
            if a_term is not None and a_dims != lev_dims:
                return '%s(term %s) must depend on the same single dimension as %s' \
                       % (a_name, a_term, vname)
            if b_dims != lev_dims:
                return '%s(term b) must depend on the same single dimension as %s' \
                       % (b_name, vname)

            # 2-3-dimensional ps
            if ps.ndim not in (2,3):
                return '%s(term ps) variable must be 2- or 3-dimensional' % ps_name

            if a_term is not None:
                # lev = a+b
                res = abs(lev - (a+b))

                if not (res < eps).all():
                    return '%s != %s(term a) + %s(term b) (see appendix D)' % (vname,a_name, b_name)
            return ''

        # generic check function for other standard_names
        def generic_constraints(ncfile, vname, terms_mapping, ndims={}, scalars=()):
            '''
            generic check function,
            only tries to check if dimensions of variables are correct

            the ndims parameter will have to be set by a wrapper according to the standard_name the check
            should be applied to. it is a dict linking formula_term and required dimensions (tuples)

            scalars is an optional tuple that enforces scalar constraints on listed terms.
            scalars do not have to be listed in ndims, but every term has to appear either
            in the one or in the other

            it is expected that all terms listed in ndims or scalars are also present in the terms_mapping
            (this should have been checked before already)
            '''
            # check everyting from ndims (non-scalars)
            for term, dims in ndims.iteritems():
                term_name = terms_mapping[term]
                data = ncfile.variables[term_name][:]
                if not data.ndim in dims:
                    return '%s(term %s) variable must be %s-dimensional' \
                            % (term_name, term, '- or '.join([str(i) for i in dims]))

            # check scalars
            for term in scalars:
                term_name = terms_mapping[term]
                data = ncfile.variables[term_name][:]
                if not (data.ndim == 0 or (data.ndim, len(data)) == (1,1)):
                    return '%s(term %s) must be scalar or 1-dimensional with length 1' % (term_name, term)

        constraints = { 'atmosphere_ln_pressure_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'lev':(1,)}, scalars=('p0',)),
                        'atmosphere_sigma_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'sigma':(1,), 'ps':(2,3)}, scalars=('ptop',)),
                        'atmosphere_hybrid_sigma_pressure_coordinate':
                            hbsigma_constraints,
                        'atmosphere_hybrid_height_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'orog':(2,3), 'a':(1,), 'b':(1,)}),
                        'atmosphere_sleve_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, scalars=('ztop',),
                                            ndims={'a':(1,), 'b1':(1,), 'b2':(1,), 'zsurf1':(2,3), 'zsurf2':(2,3)}),
                        'ocean_sigma_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'eta':(3,), 'sigma':(1,), 'depth': (2,)}),
                        'ocean_s_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'eta':(3,), 's':(1,), 'depth': (2,)},
                                                                     scalars=('a', 'b', 'depth_c')),
                        'ocean_sigma_z_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'eta':(3,), 'sigma':(1,), 'depth': (2,), 'zlev': (1,)},
                                                                     scalars=('nsigma', 'depth_c')),
                        'ocean_double_sigma_coordinate':
                            lambda a,b,c: generic_constraints(a,b,c, ndims={'sigma':(1,), 'depth': (2,)},
                                                                     scalars=('z1', 'z2', 'a', 'href')),
                        }
        return formula_terms, constraints

    @classmethod
    def _setup_grid_mappings(cls):
        '''
        generate dict of grid mapping names with their allowed attributes
        '''
        mappings = {'albers_conical_equal_area': ['standard_parallel', 'longitude_of_central_meridian',
                                                  'latitude_of_projection_origin', 'false_easting', 'false_northing'],
                    'azimuthal_equidistant': ['longitude_of_projection_origin', 'latitude_of_projection_origin',
                                              'false_easting', 'false_northing'],
                    'lambert_azimuthal_equal_area': ['longitude_of_projection_origin', 'latitude_of_projection_origin',
                                                      'false_easting', 'false_northing'],
                    'lambert_conformal_conic': ['standard_parallel', 'longitude_of_central_meridian',
                                                'latitude_of_projection_origin', 'false_easting', 'false_northing'],
                    'polar_stereographic': ['straight_vertical_longitude_from_pole', 'latitude_of_projection_origin',
                                            'standard_parallel', 'scale_factor_at_projection_origin', 'false_easting',
                                            'false_northing'],
                    'rotated_latitude_longitude': ['grid_north_pole_latitude', 'grid_north_pole_longitude',
                                                   'north_pole_grid_longitude'],
                    'stereographic': ['longitude_of_projection_origin', 'latitude_of_projection_origin',
                                      'scale_factor_at_projection_origin', 'false_easting', 'false_northing'],
                    'transverse_mercator': ['scale_factor_at_central_meridian', 'longitude_of_central_meridian',
                                            'latitude_of_projection_origin', 'false_easting', 'false_northing']
                   }
        return mappings

    @classmethod
    def _setup_grid_mapping_constraints(cls):
        '''
        generate dict of constraints for grid mapping attributes

        returns: constraints dict

        constraints include len(scalar or vector), ranges and data types so far but
        are meant to be extensible by adding more items to the lists if needed

        while not always explicitly stating the data type of the attributes, it is clear from the
        descriptions (see "Table F.1. Grid Mapping Attributes" in CF-1.0) which data type is required
        this function lists all the attributes from Table F.1 for completeness' sake,
        even if grid_mapping_name is acutally checked somewhere else as it serves a special purpose

        the first list item specifies the data type of the attribute in form of a one-character string
        that is either 'N' (numerical value) or 'S' (string value). As stated above, all bounded
        attributes have to be of type 'N', even if it is not explicitly mentioned in CF-1.0

        the second list item contains a tuple in form of (l_bnd, u_bnd) if data type is 'N' and is
        unused otherwise

        the third list item contains a tuple of the same form meaning the number
        of items to be expected in the attribute (min, max), this also only applies to type 'N' attributes

        items 2 and 3 should not be evaluated if type is 'S' and could thus also be missing, but are set
        to None here to be more explicit about it

        returns: constraints dict
        '''
        eps = cls.epsilon
        inf = float('inf')
        constraints = {'false_easting':                         ['N', (-inf, inf), (1,1)],
                       'false_northing':                        ['N', (-inf, inf), (1,1)],
                       'grid_mapping_name':                     ['S', None, None],
                       'grid_north_pole_latitude':              ['N', (-inf, inf), (1,1)],
                       'grid_north_pole_longitude':             ['N', (-inf, inf), (1,1)],
                       'latitude_of_projection_origin':         ['N', (-90.0, 90.0), (1,1)],
                       'longitude_of_central_meridian':         ['N', (-180.0, 180.0-eps), (1,1)],
                       'longitude_of_projection_origin':        ['N', (-180.0, 180.0-eps), (1,1)],
                       'north_pole_grid_longitude':             ['N', (-inf, inf), (1,1)],
                       'scale_factor_at_central_meridian':      ['N', (0.0+eps, inf), (1,1)],
                       'scale_factor_at_projection_origin':     ['N', (0.0+eps, inf), (1,1)],
                       'standard_parallel':                     ['N', (-90.0, 90.0), (1,2)],
                       'straight_vertical_longitude_from_pole': ['N', (-180.0, 180.0-eps), (1,1)],
                      }
        return constraints

    def get_messagelog(self):
        '''
        get MessageLog of file processed
        '''
        return self.msg

    def _setup_run(self, filename):
        '''
        do some common tasks to prepare object for test run.
        - write some INIT log messages
        - check for and open input file
        - read some commonly used metadata from input file
        - ...

        filename: full path of file to check
        '''
        # first check if file exists
        if not os.path.isfile(filename):
            raise CFCheckerError('file %s does not exist, cannot check' % filename)

        # set filename reference object-wide
        self.filename = filename

        # clear messagelog and log some INIT messages
        self.msg.clear()
        # cfchecker version number
        self.msg.info('INIT', 'running CFchecker version %s' % __version__ )
        # log convention version that is checked
        self.msg.info('INIT', 'checking compliance with convention %s' % self.cf_convention )
        # tests to run (if not all)
        if not self.tests_to_run is self.all_tests:
            testids = [t.replace('_test_', '').replace('_', '.') for t in self.tests_to_run]
            self.msg.info('INIT', 'will only run the following tests: %s' % ', '.join(testids))

        self.msg.info('INIT', 'using standard name table version: %s, last modified: %s' %
                      (self.name_table.get_version(), self.name_table.get_last_modified()))

        # open file for reading
        # explicit format in case we don't have a proper extension
        self.ncfile = ncdf4(self.filename, mode='r')


        # get bounds, dimension, climatology and label information from file
        self.bounds = find_bounds(self.ncfile)
        self.dim_vars = find_dim_vars(self.ncfile)
        self.auxiliary_coord_vars = find_auxiliary_coord_vars(self.ncfile)
        self.clim_vars = find_climatology_vars(self.ncfile)
        self.label_vars = find_label_vars(self.ncfile)
        self.compress_vars = find_compress_vars(self.ncfile)
        self.time_dims = map_dim_vars_to_types(self.ncfile, reverse_map=True).get('T', [])
        self.grid_mappings = find_grid_mapping_vars(self.ncfile)

    def run(self, filename):
        '''
        run all requested tests

        filename: full path of file to check
        '''
        self._setup_run(filename)

        for test_name in self.tests_to_run:
            getattr(self, test_name)()

        return
