'''
CF-1.4 changes based on CF-1.3 checker code

implementation of requirements and recommendations from
http://cf-pcmdi.llnl.gov/conformance/requirements-and-recommendations/1.4/ and
http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/cf-conventions.html

this code is an extension of the CF-1.0 checker base code in the cf10 module
and the modifications by the CF-1.1, CF-1.2 and CF-1.3 checker codes in the
cf11, cf12 and cf13 modules respectively
'''
import re, os

from utils import is_string_type, flatten, find_duplicates, find_auxiliary_coord_vars, \
                           find_scalar_coord_vars, map_dim_vars_to_types, find_area_type_vars, \
                           parse_cell_methods, parse_cell_methods_where_statement, \
                           parse_cell_methods_climatology_statement, units_known, \
                           is_monotonic

from errors import CellMethodsParseError

from CFchecker import cf13, _datadir
from CFchecker.area_type_table import AreaTypeTable

class CF14Tests(cf13.CF13Tests):
    '''
    changes compared to 1.3: modified tests 5, 6.1 and 7.3 (see below for more information)
    '''
    cf_convention = 'CF-1.4'

    def __init__(self, name_table_file=os.path.join(_datadir, 'cf-standard-name-table.xml'),
                 area_type_file=os.path.join(_datadir, 'area-type-table.xml'),not_implemented_info=False,
                 max_entries=500, tests_to_run=None):

        cf13.CF13Tests.__init__(self, name_table_file=name_table_file ,not_implemented_info=not_implemented_info,
                                max_entries=max_entries,tests_to_run=tests_to_run)

        # initizalize area type table before use
        self.area_type_table = self._load_area_type_table(area_type_file)

    def _test_5(self):
        '''
        5 Coordinate Systems
        Requirements:
        * All of a variable's dimensions that are latitude, longitude, vertical, or time dimensions must
          have corresponding coordinate variables. (no idea how to check that)
        * A coordinate variable must have values that are strictly monotonic (increasing or decreasing).
        * A coordinate variable must not have the _FillValue or missing_value attributes.
        * The type of the coordinates attribute is a string whose value is a blank separated list of
          variable names. All specified variable names must exist in the file.
        * The dimensions of each auxiliary coordinate must be a subset of the dimensions of the variable
          they are attached to. An exception is a label variable which will have a trailing dimension for the
          maximum string length.
        Recommendations:
        * The name of a multidimensional coordinate variable should not match the name of any of its dimensions.

        ADDITIONS: exclude area_type variables from auxiliary coordinate checks

        Notes:
        * coordinate attribute parsing is done from scratch to catch and report errors instead of ignoring them
          like with CommonUtils.cf_netcdf.find_auxiliary_coord_vars
        '''
        all_aux_coords = set()

        for vname, var in self.ncfile.variables.iteritems():
            # check all dim_vars for monotony, _FillValue/missing_value
            if vname in self.dim_vars:
                # check monotony
                vals = var[:]
                if not is_monotonic(vals):
                    self.msg.error('5', '%s: a coordinate variable must have values that are strictly monotonic'
                                        % vname)
                # check _FillValue and missing_value presence
                if '_FillValue' in var.ncattrs() or 'missing_value' in var.ncattrs():
                    self.msg.error('5', '%s: a coordinate variable must not have the _FillValue or missing_value '
                                        'attributes' % vname)
            # gather and check coordinate attributes
            coords_str = getattr(var, 'coordinates', None)
            if coords_str is not None:
                if not is_string_type(coords_str):
                    self.msg.error('5', '%s: coordinates attribute has to be of type string' % vname)
                else:
                    # extract coordinates attribute information and check
                    aux_coords = set(coords_str.split())
                    # check for variables that are referenced but do not exist in file
                    vars_no_exist = aux_coords - set(self.ncfile.variables.keys())
                    # add all existing auxiliary variables to global list
                    all_aux_coords |= (aux_coords - vars_no_exist)
                    if len(vars_no_exist) > 0:
                        self.msg.error('5', '%s: following variables that are referenced in coordinates attribute '
                                            'do not exist: %s' % (vname, ','.join(vars_no_exist)))
                    # check if dimensions of every referenced coordinate variable are subset of data variable
                    # coordinate variables that could not be found are excluded from this test
                    var_dims = set(var.dimensions)
                    for aux_name in (aux_coords - vars_no_exist - set(self.label_vars) - set(self.area_type_vars)):
                        # label variables will be checked in 6.1, so they are excluded here
                        aux_dims = set(self.ncfile.variables[aux_name].dimensions)
                        if not aux_dims.issubset(var_dims):
                            self.msg.error('5', '%s: dimensions of auxiliary coordinate variable "%s" are not a '
                                                'subset of the data variable\'s dimensions' % (vname, aux_name))
        # check all auxiliary vars for matches between var name and dimension names (if var is multi-dimensional)
        for aux_name in all_aux_coords:
            var = self.ncfile.variables[aux_name]
            if var.ndim > 1:
                if aux_name in var.dimensions:
                    self.msg.warning('5', '%s: name of a multidimensional auxiliary coordinate variable should not '
                                          'match the name of any of its dimensions' % aux_name)

    def _test_6_1(self):
        '''
        6.1 Labels
        Requirements:
        * A variable of character type that is named by a coordinates attribute is a label variable.
          This variable must have two dimensions. The leading dimension (CDL order) must match one
          of those of the associated variable, and the trailing dimension is for the maximum string length.

        CF-1.4 addition: exclude variables from test that have a standard_name of "area_type"
        '''
        for vname, var in self.ncfile.variables.iteritems():
            coordinates = getattr(var ,'coordinates', None)
            if coordinates is not None and is_string_type(coordinates):
                # only check those coordinate variables that are actually there
                cnames = [cname for cname in coordinates.split() if cname in self.ncfile.variables]
                for cname in cnames:
                    coord_var = self.ncfile.variables[cname]
                    std_name = getattr(coord_var, 'standard_name', None)
                    if is_string_type(std_name) and std_name == 'area_type':
                        # variables with a standard_name of area_type are skipped (see above)
                        continue
                    if coord_var.dtype.char == 'S':
                        # variable is string and referenced by coordinates attribute
                        # according to CF-1.0 this means it should be a label variable
                        if not coord_var.ndim == 2:
                            self.msg.error('6.1', '%s: label variables must have exactly 2 dimensions, not %d'
                                                   % (cname, coord_var.ndim))
                        elif not coord_var.dimensions[0] in var.dimensions:
                            # first dimension of coordinate variable has to be from referencing variable's dimensions
                            self.msg.error('6.1', '%s: leading dimension (CDL order) must match one of those of '
                                                  'the referencing variable' % cname)

    def _test_7_3(self):
        '''
        7.3 Cell Methods
        Requirements:
        * The type of the cell_methods attribute is a string whose value is one or more blank separated
          word lists, each with the form
           dim1: [dim2: [dim3: ...]] method [where type1 [over type2]] [within|over days|years] [(comment)]
          where brackets indicate optional words. The valid values for dim1 [dim2 [dim3 ...] ] are the names
          of dimensions of the data variable, names of scalar coordinate variables of the data variable,
          valid standard names, or the word area.
          The valid values of method are contained in Appendix E. The valid values for type1 are the name
          of a string-valued auxiliary or scalar coordinate variable with a standard_name of area_type,
          or any string value allowed for a variable of standard_name of area_type.
          If type2 is a string-valued auxiliary coordinate variable, it is not allowed to have a leading
          dimension (the number of strings) of more than one. When the method refers to a climatological
          time axis, the suffixes for within and over may be appended.
         * A given dimension name may only occur once in a cell_methods string.
           An exception is a climatological time dimension.
         * The comment, if present, must take the form
            ([interval: value unit [interval: ...] comment:] remainder)
           The remainder text is not standardized. If no interval clauses are present, the entire comment
           is therefore not standardized. There may be zero interval clauses, one interval clause,
           or exactly as many interval clauses as there are dims to which the method applies.
           The value must be a valid number and the unit a string that is recognizable by the udunits package.
         Recommendations:
         * If a data variable has any dimensions or scalar coordinate variables referring to horizontal,
           vertical or time dimensions, it should have a cell_methods attribute with an entry for each of
           these spatiotemporal dimensions or scalar coordinate variables.
           (The horizontal dimensions may be covered by an area entry.)
         * Except for entries whose cell method is point, all numeric coordinate variables and scalar
           coordinate variables named by cell_methods should have bounds or climatology attributes.

        NOTE:
        type1 and type2 will not be checked closer because criteria are not clear at this point,
        so any valid cf_name will be accepted (already checked by parse_cell_methods)
        '''
        re_clim_modifier = re.compile('(over|within) (years|days)$')
        re_num = re.compile('^(\d+(\.\d*)?)$')
        re_intervals_and_comment = re.compile('^(interval: \S+ \w+ ?)+(comment: ?(.*))?$')
        scalar_coord_vars = find_scalar_coord_vars(self.ncfile)
        var_type_map = map_dim_vars_to_types(self.ncfile)
        type_std_names = {'X': set(['longitude']),
                          'Y': set(['latitude']),
                          'Z': set(['air_pressure', 'atmosphere_hybrid_sigma_pressure_coordinate']),
                          'T': set(['time']) }
        types_to_names = {'X': 'horizontal',
                          'Y': 'horizontal',
                          'Z': 'vertical',
                          'T': 'time'}
        for vname, var in self.ncfile.variables.iteritems():
            cell_methods = getattr(var, 'cell_methods', None)
            if cell_methods is not None:
                #cell_methods found
                if not is_string_type(cell_methods):
                    self.msg.error('7.3', '%s: cell_methods has to be of type string' % vname)
                else:
                    try:
                        cell_methods = parse_cell_methods(cell_methods)
                    except CellMethodsParseError, err:
                        self.msg.error('7.3', '%s: cell_methods could not be parsed: %s' % (vname, str(err)))
                    else:
                        if len(cell_methods) == 0:
                            self.msg.error('7.3', '%s: cell_methods attribute given but no content found' % vname)
                        else:
                            # extract all dimensions from parsed cell_methods
                            all_dims = flatten([elem[0] for elem in cell_methods])
                            # find climatological time dimensions
                            clim_dims = set(flatten([elem[0] for elem in cell_methods if re_clim_modifier.match(elem[2])]))
                            # find duplicate dimension names and complain if it is not a climatological time dimension
                            dupe_dims = find_duplicates(all_dims)
                            for dupe_dim in dupe_dims:
                                if not dupe_dim in clim_dims:
                                    self.msg.error('7.3', '%s: dimension name "%s" is only allowed to occur once in '
                                                          'cell_methods string' % (vname, dupe_dim))
                            # check if all specified dimensions are associated with the variable, are
                            # known standard_names, the word "area" or associated scalar coordinate variables
                            for dim in set(all_dims):
                                if not (dim in var.dimensions or self.name_table.exists(dim) or dim == 'area' or dim in scalar_coord_vars.get(vname, [])):
                                    self.msg.error('7.3', '%s: cell_methods dimension "%s" has to be a dimension '
                                                          'of the associated variable, a valid standard name, the word "area" '
                                                          'or an associated scalar coordinate variable'
                                                          % (vname, dim))
                            # check the optional where ... over ... modifier for valid types
                            for cell_method in cell_methods:
                                type1, type2 = parse_cell_methods_where_statement(cell_method[2])
                                for idx, typ in enumerate((type1, type2)):
                                    if typ is None:
                                        # first occurence of None can break the loop: if type1 is missing, there can be no type2
                                        break
                                    if not self._check_where_type(vname, typ, type2check=(idx==1)):
                                        self.msg.error('7.3', '%s: cell_methods where type%d "%s" must be a valid variable name or area type.'
                                                               % (vname, idx+1, typ))

                            # check if all dimensions and scalar coordinate variables that
                            # represent X,Y,Z,T dimensions have a cell_methods entry
                            # first check dimensions
                            for dim in set(var.dimensions): # use set to get unique dimension names in case of multiple dependencies
                                var_type = var_type_map.get(dim)
                                if var_type in ('X','Y'):
                                    if not (dim in all_dims or 'area' in all_dims or
                                            len(set(all_dims) & type_std_names[var_type]) > 0):
                                        self.msg.warning('7.3', '%s: variable dimension "%s" refers to %s dimension, '
                                                                'it should have a cell_methods entry covering this dimension'
                                                                % (vname, dim, types_to_names[var_type]))
                                elif var_type in ('Z', 'T'):
                                    if not (dim in all_dims or len(set(all_dims) & type_std_names[var_type]) > 0):
                                        self.msg.warning('7.3', '%s: variable dimension "%s" refers to %s dimension, '
                                                                'it should have a cell_methods entry covering this dimension'
                                                                % (vname, dim, types_to_names[var_type]))
                            # then check scalar coordinate variables
                            for coord in scalar_coord_vars.get(vname, []):
                                coord_var = self.ncfile.variables[coord]
                                # try to find out type of variable, this detection could probably be improved
                                from_units = self._deduce_coord_type(getattr(coord_var, 'units', None), getattr(coord_var, 'positive', None))
                                from_std_name = None
                                std_name = getattr(var, 'standard_name', None)
                                for var_type in var_type_map:
                                    if var_type_map[var_type] is not None and std_name in var_type_map[var_type]:
                                        from_std_name = var_type
                                        break
                                if from_units is not None and from_units == from_std_name:
                                    var_type = from_units
                                    # found guess for X,Y,Z,T
                                    if coord not in all_dims or not (var_type in ('X','Y') and 'area' in all_dims):
                                        self.msg.warning('7.3', '%s: scalar coordinate variable "%s" refers to %s dimension, '
                                                                'it should have a cell_methods entry covering this dimension'
                                                                % (vname, coord, types_to_names[var_type]))
                            # check if all dimension variables abd scalar coordinate variables whose method is not point
                            # have either bounds or climatology attributes
                            # variables that are not found in the file will be skipped silently, they should have been reported before
                            to_check = flatten([item[0] for item in cell_methods if item[1] != 'point'])
                            for dim_vname in to_check:
                                if not (dim_vname in self.ncfile.dimensions or dim_vname in scalar_coord_vars.get(vname, [])) or dim_vname not in self.ncfile.variables:
                                    # skip all variables that have neither a dimension named after them (those are not coordinate variables)
                                    # nor are scalar coordinate variables, and of course exclude all variables that do not exist at all
                                    continue
                                elif not set(self.ncfile.variables[dim_vname].ncattrs()) & set(['bounds','climatology']):
                                    # variable has no bounds or climatology attribute
                                    self.msg.warning('7.3', '%s: coordinate/scalar variable "%s" named by cell_methods should have '
                                                            'either bounds or climatology attribute' % (vname, dim_vname))
                            # call outsourced comment check
                            self._check_cell_methods_comments(vname, cell_methods)

    def _check_where_type(self, vname, value, type2check=False):
        '''
        check the validity of the type1 and type2 values in the where ... over ... cell_methods clause

        from Requirements and Recommendations 7.3:
        [...]
        The valid values for type1 are the name of a string-valued auxiliary or scalar coordinate
        variable with a standard_name of area_type, or any string value allowed for a variable of
        standard_name of area_type. If type2 is a string-valued auxiliary coordinate variable,
        it is not allowed to have a leading dimension (the number of strings) of more than one.
        [...]

        ADDITION: constraints for type2 are not entirely clear, this code assumes that type2 variables
                  can have a max. rank of 2

        vname: variable name the cell_methods are attached to
        value: value of type to check (string)
        type2check: check additional constraints for type2 (bool)
        returns: if value is a valid value for this type
        '''
        aux_coord_vars = find_auxiliary_coord_vars(self.ncfile).get(vname, [])
        is_valid = False
        if value in aux_coord_vars and self.ncfile.variables[value].dtype.char == 'S':
            aux_var = self.ncfile.variables[value]
            # string-valued aux. coord var
            std_name = getattr(aux_var, 'standard_name', None)
            if is_string_type(std_name) and std_name == 'area_type':
                # additional check in case it's a type2
                if type2check:
                    if aux_var.ndim == 2 and len(self.ncfile.dimensions[aux_var.dimensions[0]]) == 1:
                        # 2-dimensional, with leading dimension size 1
                        is_valid = True
                else:
                    is_valid = True
        elif self.area_type_table.exists(value):
            # known area type
            is_valid = True

        return is_valid

    def _load_area_type_table(self, area_type_table_file):
        area_type_table = AreaTypeTable(area_type_table_file)
        return area_type_table

    def _setup_run(self, filename):
        cf13.CF13Tests._setup_run(self, filename)

        self.msg.info('INIT', 'using area type table version: %s, date: %s' %
                      (self.area_type_table.get_version(), self.area_type_table.get_date()))
        self.area_type_vars = find_area_type_vars(self.ncfile)
