'''
iso_time module from datafed package of OWS server (as created by Kari Hoijvari)
'''

from __future__ import division
from __future__ import with_statement
__metaclass__ = type
import datetime
import time
import re
import math

class TimePeriodicity:
    year = "year"
    month = "month"
    week = "week"
    day = "day"
    hour = "hour"
    minute = "minute"
    second = "second"
    TimePeriodicities = [year, month, week, day, hour, minute, second]
    def select(self, dta):
        return [ii for ii in xrange(len(dta)) if self.has_datetime(dta[ii])]

class TimePeriodicityEmpty(TimePeriodicity):
    def __eq__(self, other):
        return self.__class__ == other.__class__
    def has_datetime(self, dt):
        return False
    def all_datetimes(self, period=None):
        return []

class TimePeriodicityScalar(TimePeriodicity):
    def __init__(self,dt):
        if not isinstance(dt, datetime.datetime):
            raise AssertionError("TimePeriodicityScalar expects datetime.datetime, not " + repr(dt))
        self.dt = dt
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.dt == other.dt
    def has_datetime(self, dt):
        return self.dt == dt
    def all_datetimes(self, period=None):
        return [self.dt]

class TimePeriodicityRange(TimePeriodicity):
    def __init__(self,time_min,time_max):
        if not isinstance(time_min, datetime.datetime):
            raise AssertionError("TimePeriodicityRange time_min must be datetime.datetime, not " + repr(time_min))
        if not isinstance(time_max, datetime.datetime):
            raise AssertionError("TimePeriodicityRange time_max must be datetime.datetime, not " + repr(time_max))
        self.time_min = time_min
        self.time_max = time_max
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return (self.time_min == other.time_min
                and self.time_max == other.time_max)
    def all_datetimes(self, period=(TimePeriodicity.day, 1)):
        return TimePeriodicityComputed(period[0], period[1], self.time_min, self.time_max).all_datetimes()
    def get_range(self):
        return [self.time_min, self.time_max]

class TimePeriodicityComputed(TimePeriodicity):
    def __init__(self,unit,mul,time_min,time_max):
        if not unit in TimePeriodicity.TimePeriodicities:
            raise AssertionError("Unit must be one of %s, not %s" % (repr(TimePeriodicity.TimePeriodicities), repr(unit)))
        if not isinstance(time_min, datetime.datetime):
            raise AssertionError("TimePeriodicityRange time_min must be datetime.datetime, not " + repr(time_min))
        if not isinstance(time_max, datetime.datetime):
            raise AssertionError("TimePeriodicityRange time_max must be datetime.datetime, not " + repr(time_max))
        self.unit = unit
        self.mul = int(mul)
        self.time_min = time_min
        # "round" self.time_max to be consistent with periodicity and <= input time_max
        self.time_max = self.index2time(self.time2index(time_max))
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return (self.unit == other.unit
                and self.mul == other.mul
                and self.time_min == other.time_min
                and self.time_max == other.time_max)
    def has_datetime(self, dt):
        idx = self.time2index(dt)
        exact = self.index2time(idx)
        return exact == dt
    def time2index(self, index_time):
        return time2index(index_time, self.unit, self.mul, self.time_min)
    def index2time(self, index_time):
        return index2time(index_time, self.unit, self.mul, self.time_min)
    def all_datetimes(self, period=None):
        ret = []
        itime = 0
        idx_last = self.time2index(self.time_max)
        dt = self.index2time(itime)
        for itime in xrange(idx_last+1):
            ret.append(self.index2time(itime))
        return ret
    def duration(self):
        return format_duration(self.unit, self.mul)
    def get_range(self):
        return [self.time_min, self.time_max]

def format_duration(tp_unit, tp_mul):
    sep = ""
    unit = ""
    if tp_unit == TimePeriodicity.year:
        unit = "Y"
    elif tp_unit == TimePeriodicity.month:
        unit = "M"
    elif tp_unit == TimePeriodicity.day:
        unit = "D"
    elif tp_unit == TimePeriodicity.hour:
        sep = "T"
        unit = "H"
    elif tp_unit == TimePeriodicity.minute:
        sep = "T"
        unit = "M"
    elif tp_unit == TimePeriodicity.second:
        sep = "T"
        unit = "S"
    return 'P%s%d%s' % (sep, tp_mul, unit)

def csmod(n,d):
    if n >= 0:
        return int(round(n % d))
    else:
        return int(round(-(abs(n)%d)))
def csdiv(n,d):
    if n >= 0:
        return int(round(n // d))
    else:
        return int(round(-(abs(n)//d)))
def days2timedelta(days):
    return datetime.timedelta(days)
def hours2timedelta(hours):
    return datetime.timedelta(0,0,0,0,0,hours)
def minutes2timedelta(minutes):
    return datetime.timedelta(0,0,0,0,minutes)
def seconds2timedelta(seconds):
    return datetime.timedelta(0,seconds)
def timedelta_hours(td):
    return timedelta_minutes(td)/60
def timedelta_minutes(td):
    return timedelta_seconds(td)/60
def timedelta_seconds(td):
    return 3600*24*td.days + td.seconds + td.microseconds/1000000

def time2index(index_time, unit, mul, time_min):
    if unit == TimePeriodicity.year:
        ret = year2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.month:
        ret = month2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.week:
        ret = week2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.day:
        ret = day2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.hour:
        ret = hour2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.minute:
        ret = minute2index(mul, time_min, index_time)
    elif unit == TimePeriodicity.second:
        ret = second2index(mul, time_min, index_time)
    else:
        raise AssertionError("wrong periodicity_unit " + unit)
    return ret
def index2time(index, unit, mul, time_min):
    if unit == TimePeriodicity.year:
        ret = index2year(mul, time_min, index)
    elif unit == TimePeriodicity.month:
        ret = index2month(mul, time_min, index)
    elif unit == TimePeriodicity.week:
        ret = index2week(mul, time_min, index)
    elif unit == TimePeriodicity.day:
        ret = index2day(mul, time_min, index)
    elif unit == TimePeriodicity.hour:
        ret = index2hour(mul, time_min, index)
    elif unit == TimePeriodicity.minute:
        ret = index2minute(mul, time_min, index)
    elif unit == TimePeriodicity.second:
        ret = index2second(mul, time_min, index)
    else:
        raise AssertionError("wrong periodicity_unit " + unit)
    return ret

def year2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    return int(round((index_time.year - time_min.year) // mul))
def index2year(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul * index
    return datetime.datetime(time_min.year + index, 1, 1)
def month2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    years = index_time.year - time_min.year
    months = index_time.month - time_min.month
    return csdiv(years * 12 + months, mul)
def index2month(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul * index
    month = time_min.month-1 + index
    if month >= 0:
        y = time_min.year + month // 12
        m = month % 12 + 1
        ret = datetime.datetime(y, m, 1)
    else:
        y = time_min.year -(abs(month + 1)//12) -1
        m = 12 -(abs(month + 1)%12)
        ret = datetime.datetime(y, m, 1)
    return ret
def week2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    td = index_time - time_min
    return csdiv(csdiv(td.days, 7), mul)
def index2week(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul*index
    return time_min + datetime.timedelta(7 * index)
def day2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    td = index_time - time_min
    days = td.days
    return csdiv(days, mul)
def index2day(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul*index
    return time_min + datetime.timedelta(index)
def hour2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    td = index_time - time_min
    return int(round(timedelta_hours(td) // mul))
def index2hour(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul*index
    return time_min + hours2timedelta(index)
def minute2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    td = index_time - time_min
    tm = int(math.floor(timedelta_minutes(td)))
    return int(round(tm // mul))
def index2minute(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul*index
    return time_min + minutes2timedelta(index)
def second2index(mul, time_min, index_time):
    assert mul > 0, "Periodicity multiplier must be positive."
    td = index_time - time_min
    tm = int(math.floor(timedelta_seconds(td)))
    return tm // mul
def index2second(mul, time_min, index):
    assert mul > 0, "Periodicity multiplier must be positive."
    index = mul*index
    return time_min + seconds2timedelta(index)

class TimePeriodicityEnumerated(TimePeriodicity):
    def __init__(self,times):
        self.times = times
        self.time_min = min(times)
        self.time_max = max(times)
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return false
        return self.times == other.times
    def has_datetime(self, dt):
        for sdt in self.times:
            if sdt.has_datetime(dt) > 0:
                return True
        return False
    def all_datetimes(self, period=None):
        ret = []
        for dt in self.times:
            if dt.__class__ == datetime.datetime:
                ret.append(dt)
            else:
                ret.extend(dt.all_datetimes(period))
        return ret
    def get_range(self):
        return [self.time_min, self.time_max]

def parse(text):
    if text.find(',') >= 0:
        return parse_enumerated(text)
    elif text.find('/') >= 0:
        return parse_range_or_now(text)
    else:
        return TimePeriodicityScalar(parse_one_datetime(text))

# regex to match ISO 8601 format
# current issues that violate total ISO compliance (list probably incomplete):
#   - delimiters (- and :) are mandatory
#   - does not enforce zero-padding of month, day, hour, minute, second
#   - does not accept timezone information, only "Z"
re_iso_time_format = re.compile('^(?P<year>\d\d\d\d)(-(?P<month>[01]?\d)'
                                '(-(?P<day>[0-3]?\d)(( |T)(?P<hour>[0-2]?\d)'
                                '(:(?P<minute>[0-6]?\d)(:(?P<second>[0-6]?\d)'
                                '(.(?P<usec>\d+))?)?)?Z?)?)?)?$')
def parse_one_datetime(text):
    # default values to use for creating datetime objects when element was not specified
    datetime_defaults = {'month': 1,
                         'day': 1,
                         'hour': 0,
                         'minute': 0,
                         'second': 0,
                         'usec': 0,
                         }

    text = text.strip()

    iso_time_match = re_iso_time_format.match(text)

    if iso_time_match is None:
        raise Exception('Cannot parse "%s" as ISO time format' % text)

    matches = iso_time_match.groupdict()

    dt_vals = {}

    # no default needed for year, it always has to be there
    dt_vals['year'] = int(matches['year'])

    # process all the other elements and use defaults if no data was given
    for item in ('month', 'day', 'hour', 'minute', 'second'):
        if matches[item] is None:
            dt_vals[item] = datetime_defaults[item]
        else:
            dt_vals[item] = int(matches[item])

    # usec needs special treatment for proper representation
    if matches['usec'] is None:
        dt_vals['usec'] = datetime_defaults['usec']
    else:
        dt_vals['usec'] = int((matches['usec'] + '00000')[0:6])

    return datetime.datetime(dt_vals['year'], dt_vals['month'], dt_vals['day'], dt_vals['hour'],
                             dt_vals['minute'], dt_vals['second'], dt_vals['usec'])

def parse_enumerated(iso_time_csv):
    ret = []
    for t in iso_time_csv.split(','):
        ret.append(parse(t))
    return TimePeriodicityEnumerated(ret)

def is_int(text):
    try:
        int(text)
        return True
    except ValueError:
        return False

def periodicity2timedelta(unit, mul):
    if unit == TimePeriodicity.week:
        return days2timedelta(7*mul)
    elif unit == TimePeriodicity.day:
        return days2timedelta(mul)
    elif unit == TimePeriodicity.hour:
        return hours2timedelta(mul)
    elif unit == TimePeriodicity.minute:
        return minutes2timedelta(mul)
    elif unit == TimePeriodicity.second:
        return seconds2timedelta(mul)
    else:
        raise AssertionError("wrong periodicity_unit " + unit)

def parse_periodicity(periodicity):
    if periodicity[0] != 'P':
        raise Exception("periodicity must start with P, not " + periodicity)
    state = "NX"
    mul = 0
    unit = ""
    peri_group = "date"
    for iperi in xrange(1, len(periodicity)):
        s = periodicity[iperi]
        if state == "NX":
            if is_int(s):
                state = "N"
                mul = int(s)
            else:
                if s == "T":
                    peri_group = "time"
                else:
                    raise Exception("illegal separator " + s)
        elif state == "N":
            if is_int(s):
                mul = 10*mul + int(s)
            else:
                if s == "Y":
                    unit = TimePeriodicity.year
                    if peri_group != "date":
                        raise Exception("wrong T")
                    break
                elif s == "M":
                    if peri_group == "date":
                        unit = TimePeriodicity.month
                    else:
                        unit = TimePeriodicity.minute
                        if peri_group != "time":
                            raise Exception("wrong T")
                elif s == "D":
                    unit = TimePeriodicity.day
                    if peri_group != "date":
                        raise Exception("wrong T")
                elif s == "W":
                    unit = TimePeriodicity.week
                    if peri_group != "date":
                        raise Exception("wrong T")
                elif s == "H":
                    unit = TimePeriodicity.hour
                    if peri_group != "time":
                        raise Exception("wrong T")
                elif s == "S":
                    unit = TimePeriodicity.second
                    if peri_group != "time":
                        raise Exception("wrong T")
                else:
                    raise Exception("illegal periodicity " + s)
                state = "end"
        else:
            raise AssertionError("not yet implemented periodicity parsing for " + periodicity)
    return unit, mul

def parse_range_or_now(range_time):
    parts = range_time.split('/')
    if not len(parts) in [2, 3]:
        raise Exception("date range must have 2 or 3 parts, not %i" % len(parts))

    time_min = parse(parts[0])
    time_max = parse(parts[1])
    if time_min == time_max:
        if len(parts) == 2:
            return time_min
        else:
            # ISO 8601 oddidity: same min/max with periodicity results in 'now'
            # Section D.5 Examples Page 51 in document OGC 04-024
            # EXAMPLE 4 Current data refreshed every 30 min (periodic interval):
            # 2000-06-18T14:30Z/2000-06-18T14:30Z/PT30M
            return TimePeriodicityScalar(datetime.datetime.now())

    if len(parts) == 2:
        return TimePeriodicityRange(time_min.dt, time_max.dt)

    assert len(parts) == 3
    unit, mul = parse_periodicity(parts[2])
    return TimePeriodicityComputed(unit, mul, time_min.dt, time_max.dt)

def to_str(dt):
    return dt.isoformat('T')
def to_strz(dt):
    return dt.isoformat('T') + 'Z'

def day_of_year(dt):
    orig = datetime.datetime(dt.year, 1, 1)
    span = dt-orig
    return int(math.floor(span.days+1))

def encode_starttime(units):
    parts = units.split(' ', 2)
    assert len(parts) == 3
    unit = parts[0]
    assert unit[-1] == 's'
    assert parts[1] == 'since'
    return parse_one_datetime(parts[2]), unit[:-1]

def is_linear(stamps):
    step = stamps[1] - stamps[0]
    for ii in xrange(len(stamps)-1):
        if step != stamps[ii+1] - stamps[ii]:
            return False
    return True

def from_enum(units, stamps):
    time_min, unit = encode_starttime(units)
    if len(stamps) == 0:
        return TimePeriodicityEmpty()
    elif len(stamps) == 1:
        return TimePeriodicityScalar(index2time(stamps[0], unit, 1, time_min))
    else:
        if is_linear(stamps):
            step = stamps[1] - stamps[0]
            if (step >= 1) and (step % 1 == 0):
                return from_linear_enum(unit, step, time_min, stamps[0], stamps[-1])
        return from_enumerated(time_min, unit, stamps)

def from_linear_enum(unit, step, time_min, stamp_min, stamp_max):
    # sanity check. stamp_max - stamp_min should be a multiple of step
    if not (stamp_max - stamp_min) % step == 0: raise AssertionError('time indices not consistent with step size')
    if unit == 'millisecond':
        unit_change_factor = 1000
        unit = 'second'
    else:
        unit_change_factor = 1
    time_start = index2time(stamp_min // unit_change_factor, unit, 1, time_min)
    time_end = index2time(stamp_max // unit_change_factor, unit, 1, time_min)
    return TimePeriodicityComputed(unit, step // unit_change_factor, time_start, time_end)

def from_enumerated(time_min, unit, stamps):
    return TimePeriodicityEnumerated([index2time(ii, unit, 1, time_min) for ii in stamps])
