"""
utils module

dependencies:
 - cfunits (https://code.google.com/p/cfunits-python/)
    requires Unidata UDUNITS2 C Library
"""
import re, os

import numpy as np

from cfunits import Units as udunits

from CFchecker.errors import CellMethodsParseError, UnknownUnitsError

def list_files(item):
    '''
    wrapper for files and directories that allows to handle single
    file names and directories the same way.
    if item is a file: return the filename
    if item is a path: use os.walk() on it to return all filenames

    item: path of file or directory (string)
    returns: rootdir, dirs, files tuple (like os.walk)
    '''
    if os.path.isfile(item):
        yield os.path.dirname(item), [], [os.path.basename(item)]
    elif os.path.isdir(item):
        for root, dirs, files in os.walk(item):
            yield root, dirs, files
    else:
        raise OSError('supplied item %s does not exist or is neither file nor directory!' % item)

def is_string_type(item):
    '''
    determine if the supplied item can be used like a string

    item: item to test
    returns: value can be used like a string or not (bool)
    '''
    try:
        res = item + 'foo'
    except TypeError, e:
        return False
    else:
        return True

def is_num_type(item):
    '''
    determine if the supplied item can be treated like a number

    item: item to test
    returns: value can be treated like a number or not (bool)
    '''
    # only atomic types should check positive,
    # so reject all enum types directly
    if (hasattr(item, 'ndim') and item.ndim > 1) or \
       (hasattr(item, 'shape') and item.shape not in ((),(1,))) or \
       (not hasattr(item, 'ndim') and hasattr(item,'__len__')):
        return False
    try:
        res = item + 42
    except TypeError, e:
        return False
    else:
        return True

def flatten(l, ltypes=(list, tuple)):
    '''
    flatten lists (or tuples) down to a single list containing all elements
    source: http://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html

    l: input structure to flatten
    ltypes: types of structures to consider
    returns: flattened list conataining all input elements, return type equals type of l
    '''
    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            if not l[i]:
                l.pop(i)
                i -= 1
                break
            else:
                l[i:i + 1] = l[i]
        i += 1
    return ltype(l)

def find_duplicates(items):
    '''
    check given list of items for duplicates and return a list of all duplicates found.
    if no duplicates are found, the returned list is empty.
    every duplicate name is only listed once in the return value.

    items: list of items to check for duplicates
    returns: list of all duplicate items
    '''
    ret = set()
    # the last item of the list does not have to be iterated over as it will be checked before
    for idx, item in enumerate(items[:-1]):
        if item in items[idx+1:]: ret.add(item)
    return list(ret)

def is_hashable(item):
    '''
    determine if the supplied item is hashable
    the only reliable way seems to be to actually try to
    use the item for something that requires hashable objects.
    simply checking presence of __hash__ and other methods
    seems unsufficient as to be seen with numpy.ndarray which
    is not hashable but contains all the required methods.

    item: item to check
    returns: True if hashable, False otherwise
    '''
    try:
        foo = {item: 1}
        return True
    except TypeError, err:
        # only specifically check for the "unhashable type" error
        if not str(err).startswith('unhashable type:'):
            raise
        return False

def find_dim_vars(ncfile):
    '''
    find all dimension variables

    dimension variables are variables that only depend on a dimension of the same name.

    ncfile: open input netcdf file

    returns: list of dimension variable names
    '''
    ret = []
    for dim in ncfile.dimensions:
        if dim in ncfile.variables:
            if ncfile.variables[dim].dimensions == (dim,):
                ret.append(dim)
    return ret

def map_axes(ncfile, reverse_map=False):
    '''
    search all dimension variables for the axis attribute and create an entry
    with the axis name as key and dimension name as value in the return dictionary.
    non-string axis values are skipped silently

    ncfile: file handle of netcdf file
    reverse_map: bool value for map direction; False: Axis->Dimname, True:Dimname->Axis
    returns: mapping: axis name -> dimension names (list) if reverse_map == False,
                               dimension name -> axis name if reverse_map == True
    '''
    ret = {}
    axes = ['X', 'Y', 'Z', 'T']
    for dim in find_dim_vars(ncfile):
        axis = getattr(ncfile.variables[dim], 'axis', '')
        if not is_string_type(axis):
            # skip non-string axis values
            continue
        axis = axis.upper()
        if axis in axes:
            if reverse_map:
                ret[dim] = axis
            else:
                if axis in ret:
                    ret[axis].append(dim)
                else:
                    ret[axis] = [dim]
    return ret

def map_dim_vars_to_types(ncfile, reverse_map=False):
    '''
    try to determine the type of every dimension variable in the given file
    detectable so far: lat/X, lon/Y, lev/Z, time/T types

    following attributes are checked:
    - standard_name
    - axis
    - units
    - positive
    - long_name

    if there is any reason for uncertainly (like different attributes not
    pointing to the same type, no guess will be made and the variable is skipped.
    CF-1.x compatible files should always be consistent enough to be interpreted correctly.

    ncfile: open input file
    reverse_map: bool value, set to true to return a mapping from axis type (X,Y,Z,T)
                 to variable names

    returns: mapping <variable name> -> axis type if reverse_map is False,
             else: mapping <axis type> -> [variable name(s)].
             note that while variable-to-type mappings are unambiguous and thus the value will
             be a single string, type-to-variable mappings can contain more than one value per
             key and are thus packed as a list.
    '''
    var_map = {}
    # guesses for standard name attributes
    std_name_g = { 'time': 'T',
                   'longitude': 'X',
                   'latitude': 'Y',
                   'air_pressure': 'Z',
                   'altitude': 'Z',
                   'atmosphere_ln_pressure_coordinate': 'Z',
                   'atmosphere_sigma_coordinate': 'Z',
                   'atmosphere_hybrid_sigma_pressure_coordinate': 'Z',
                   'atmosphere_hybrid_height_coordinate': 'Z',
                   'atmosphere_sleve_coordinate': 'Z',
                   'ocean_sigma_coordinate': 'Z',
                   'ocean_s_coordinate': 'Z',
                   'ocean_sigma_z_coordinate': 'Z',
                   'ocean_double_sigma_coordinate': 'Z', }
    # guesses for axis attributes
    axis_g = { 'T': 'T',
               'X': 'X',
               'Y': 'Y',
               'Z': 'Z' }
    # guesses for units attribute
    units_g = { 'days since 1970-01-01': 'T',
                # will need to do another check to tell the difference between X and Y degrees
                # and to avoid false positives as UDUNITS-2 can convert between degrees and "1"
                'degrees': 'XY',
                'hPa': 'Z' }
    # guesses for lat and lon units
    x_units_g = ('degrees_east', 'degree_east', 'degrees_E', 'degree_E', 'degreesE', 'degreeE')
    y_units_g = ('degrees_north', 'degree_north', 'degrees_N', 'degree_N', 'degreesN', 'degreeN')
    # guesses for positive attribute
    positive_g = { 'up': 'Z',
                   'down': 'Z' }
    # guesses for long name attribute
    long_name_g = { 'time': 'T',
                    'longitude': 'X',
                    'latitude': 'Y',
                    'altitude': 'Z',
                    'height': 'Z',
                    'elevation': 'Z',}

    for vname in find_dim_vars(ncfile):
        var = ncfile.variables[vname]

        # first extract all attribute values
        std_name_v = getattr(var, 'standard_name', None)
        axis_v = getattr(var, 'axis', None)
        positive_v = getattr(var, 'positive', None)
        long_name_v = getattr(var, 'long_name', None)
        # then check if values are hashable and reset those to None that are not
        hash_filter = lambda item: item if is_hashable(item) else None
        std_name_v, axis_v, positive_v, long_name_v = [hash_filter(item) for item in (std_name_v, axis_v, positive_v, long_name_v)]
        # finally look up the guessed types
        std_name_t = std_name_g.get(std_name_v)
        axis_t = axis_g.get(axis_v)
        positive_t = positive_g.get(positive_v)
        long_name_t = long_name_g.get(long_name_v)

        # some more processing required for unit checking
        units = getattr(var, 'units', None)
        if not units_known(units):
            # if used units are unknown, we're done
            units_t = None
        else:
            # find out if we have guessable units compatible to the ones saved in file
            found_units = None
            for unit in units_g:
                if units_convertible(units, unit):
                    found_units = units_g[unit]
                    break
            if found_units == 'XY':
                # find out if it's X or Y
                if units in x_units_g:
                    found_units = 'X'
                elif units in y_units_g:
                    found_units = 'Y'
                else:
                    # if neither X nor Y degrees, then we have to reset found_units to nothing
                    # to exclude the false positives introduced by UDUNITS-2
                    found_units = None
            units_t = found_units

        # collect all successful partial guesses
        found_types = set([item for item in (std_name_t, axis_t, positive_t, long_name_t, units_t)
                           if item is not None])
        if len(found_types) == 1:
            # found exactly one guess for type, so save that
            var_type = found_types.pop()
        else:
            # in all other cases no definite conclusion could be reached, so no type is saved
            var_type = None
        if reverse_map:
            if var_type in var_map:
                var_map[var_type].append(vname)
            else:
                var_map[var_type] = [vname]
        else:
            var_map[vname] = var_type
    return var_map

def find_bounds(ncfile):
    '''
    search all variables for the bounds-attribute and create a mapping between the
    variable name containing the attribute and the attribute value (should be another variable name)
    ignores bounds values that point to non-existing variables

    ncfile: open input file
    returns: mapping variable name -> bounds variable (content of bounds attribute)
    '''
    ret = {}
    for vname in ncfile.variables:
        bounds = getattr(ncfile.variables[vname],'bounds', None)
        if is_string_type(bounds) and bounds in ncfile.variables:
            ret[vname] = bounds
    return ret

re_valid_cf_name = re.compile('^[a-zA-Z]\w*$')

def is_valid_cf_name(name):
    '''
    check if name follows the CF-1.0 naming rules:
    begin with a letter and be composed of letters, digits, and underscores
    (see regex above)

    name: input string to check
    returns: True if valid name, else False
    '''
    return is_string_type(name) and bool(re_valid_cf_name.match(name))

def parse_cell_methods(methods_string):
    '''
    parser for all known variations of the cell methods attribute defined in CF-1.0
    and slightly extended and clarified in CF-1.4

    syntax:
    dim1: [dim2: [dim3: ...]] method [where type1 [over type2]] [within|over days|years] \
          [([interval: value unit [interval: ...] comment:] remainder)]

    NOTES:
    this function will parse according to the syntax rules and check as many of them as it can, however, there are
    some checks that need to be done after parsing and by a function that has more information about the data at hand.
    an example is the checking for valid names. according to the specs, valid names could be dimension names
    of the corrosponding variable, standard names, a scalar coordinate variable or the world "area".
    this parser will simply accept every name that conforms with CF naming rules.
    comments are not parsed further.

    methods_string: cell_methods string to parse
    returns: list of (names, method, modifiers, comment) tuples
    '''
    # all accepted methods, case insensitive
    known_methods = ['point', 'sum', 'maximum', 'median', 'mid_range', 'minimum', 'mean', 'mode', 'standard_deviation', 'variance']
    clim_time_units = ['years', 'days']
    ret = []
    if methods_string.strip() == '':
        # return empty list for empty methods
        return ret
    # split methods string by space
    pieces = methods_string.strip().split()
    # now iterate over the pieces and try put them together in a meaningful way
    while len(pieces)>0:
        cur = pieces.pop(0) # there should be at least one piece in the list here
        names = []  # all affected names will be recorded here
        # cur piece should always end with ':'
        if not cur.endswith(':'):
            raise CellMethodsParseError('first piece does not end with ":" (\'%s\')' % cur)
        name = cur[:-1]   # remove ':'
        if not is_valid_cf_name(name):
            raise CellMethodsParseError('token does not look like a valid CF name: \'%s\'' % name)
        names.append(name)
        if len(pieces) == 0:
            # no more pieces, but processing can't be finished at this point
            raise CellMethodsParseError('input finished prematurely: method missing')
        cur = pieces.pop(0)
        # cur piece could be ending with ':' if there was no method specified for the first piece alone,
        # this means we need to get an arbitary amount of further names following the same format as first
        while cur.endswith(':'):
            name = cur[:-1]
            if not is_valid_cf_name(name):
                raise CellMethodsParseError('token does not look like a valid CF name: %s' % repr(name))
            names.append(name)
            if len(pieces) == 0:
                # there has to be more data coming
                raise CellMethodsParseError('input finished prematurely: method missing')
            cur = pieces.pop(0)
        # all names should be parsed now, so we are expecting a method
        # can't be ending on ':', so should be method
        # always convert method string to lower case as comparison should be case insensitive
        method = cur.lower()
        if not method in known_methods:
            raise CellMethodsParseError('unknown method %s' % repr(method))
        # next there could be where/over or climatology modifiers (over, within)
        modifiers = []
        if len(pieces) > 0:
            # more to process
            cur = pieces.pop(0)
            if cur in ('where',):
                # where-modifier found, type1 should follow
                where_mod = cur
                # there should be a time unit coming next
                if len(pieces) == 0:
                    raise CellMethodsParseError('input finished prematurely: where modifier type1 missing')
                cur = pieces.pop(0)
                if not is_valid_cf_name(cur):
                    raise CellMethodsParseError('type1 token does not look like a valid CF name: %s' % repr(cur))
                where_mod += ' %s' % cur
                # now there could be an optional "over" keyword coming followed by type2
                if len(pieces) > 0:
                    cur = pieces.pop(0)
                    if cur in ('over',):
                        where_mod += ' %s' % cur
                        # over modifier found, type2 should come next
                        if len(pieces) == 0:
                            raise CellMethodsParseError('input finished prematurely: where modifier type2 missing')
                        cur = pieces.pop(0)
                        if not is_valid_cf_name(cur):
                            raise CellMethodsParseError('type2 token does not look like a valid CF name: %s' % repr(cur))
                        where_mod += ' %s' % cur
                    else:
                        # did not match, put it back
                        pieces.insert(0, cur)
                modifiers.append(where_mod)
            else:
                # did not match, put it back
                pieces.insert(0, cur)
        if len(pieces) > 0:
            # still more to process
            cur = pieces.pop(0)
            # if it is either "over" or "within", it's a climatological description
            if cur in ('over', 'within'):
                climate_mod = cur
                # there should be a time unit coming next
                if len(pieces) == 0:
                    raise CellMethodsParseError('input finished prematurely: climatological time unit missing')
                cur = pieces.pop(0)
                # next piece has to be a time unit
                if not cur in clim_time_units:
                    raise CellMethodsParseError('expected climatological time unit, got "%s"' % cur)
                # append this information to the method variable
                modifiers.append('%s %s' % (climate_mod, cur))
            else:
                # did not match, put it back
                pieces.insert(0, cur)
        # merge modifiers into a single string
        modifiers = ' '.join(modifiers)
        # there could be a new name/dim coming or a comment about the method (in parentheses)
        comment = '' # empty if we don't find any
        if len(pieces) > 0:
            cur = pieces.pop(0)
            # if it is surrounded by brackets, it's extra info
            if cur.startswith('('):
                # extra info start, need to find end
                if cur.endswith(')'):
                    # end found
                    comment = cur[1:-1]    # save all the comment
                else:
                    # end not in this chunk, need to search more
                    comment = cur[1:]    # save first part of comment
                    if len(pieces) == 0:
                        raise CellMethodsParseError('incomplete comment, expected more data to come; closing parenthesis missing')
                    cur = pieces.pop(0)
                    while not cur.endswith(')'):
                        # put all the extra info chunks together until parentheses close
                        comment += (' ' + cur)
                        if len(pieces) == 0:
                            raise CellMethodsParseError('incomplete comment, expected more data to come; closing parenthesis missing')
                        cur = pieces.pop(0)
                    # fell out of loop, so current piece must end with ')'
                    comment += (' ' + cur[:-1])
                    if '(' in comment:
                        # parentheses should not be allowed in comment (closing ones will be considered end of comment,
                        # so we only have to worry about opening ones)
                        raise CellMethodsParseError('opening parentheses not allowed in comment')
            else:
                # could be first piece for next iteration, so we put it back
                pieces.insert(0, cur)
        # add parse result to result list
        ret.append((names, method, modifiers, comment.strip()))
    # loop ended, all pieces should be processed now
    return ret

def parse_cell_methods_where_statement(statement):
    '''
    parse the optional where statement types out of a single cell_methods area-climatology element
    (index 2 in tuple returned by parse_cell_methods).
    as the checks in parse_cell_methods have already run, it is guaranteed that the
    given method string has a valid syntax.

    statement: single area-climatology element (index 2) as returned by parse_cell_methods
    returns: (type1, type2); None for types that are not available
    '''
    re_clim_modifier = re.compile('(over|within) (years|days)$')
    type1, type2 = (None, None)
    # only need to do something if there is a "where" in the beginning
    if not statement.startswith('where'):
        return (None, None)

    splt = statement.split()
    # remove possible climatology modifier at the end of the list to exclude
    # the ambiguous "over"
    if re_clim_modifier.search(statement):
        splt = splt[:-2]    # remove last 2 elements which are climatology modifier
    # if there is a where, there is always a type1
    type1 = splt[1]
    # if the split is 4 elements long, then there is a type2 as well
    if len(splt) == 4:
        type2 = splt[3]

    return (type1, type2)

def parse_cell_methods_climatology_statement(statement):
    '''
    parse the optional climatology modifier and units out of a single cell_methods area-climatology element
    (index 2 in tuple returned by parse_cell_methods).
    as the checks in parse_cell_methods have already run, it is guaranteed that the
    given method string has a valid syntax.

    method_mod: single area-climatology element (index 2) as returned by parse_cell_methods
    returns: (modifier, units); (None, None) if there is no climatology modifier
    '''
    re_clim_modifier = re.compile('(over|within) (years|days)$')
    res = re_clim_modifier.search(statement)
    if res:
        return res.groups()
    else:
        return (None, None)

def find_climatology_vars(ncfile):
    '''
    try to find out which variables are climatology variables
    from the CF-1.0 standard 7.4 :
    "Climatological variables have a climatological time axis.[...]
    [...] a climatological time coordinate variable does not have a bounds attribute.
    Instead, it has a climatology  attribute [...]
    [...] time coordinates should also be recognised as climatological if they have a units
    attribute of time-units relative to midnight on 1 January in year 0 i.e. since 0-1-1 in udunits syntax [...]
    We do not recommend this convention [...]
    [...] the statistical operations are recorded in the cell_methods  attribute, which may have two or three
    entries for the climatological time dimension. [...]
    Valid values of the cell_methods attribute must be in one of the forms from the following list [...]

    time: method1 within years time: method2 over years
    [...]
    time: method1 within days time: method2 over days
    [...]
    time: method1 within days time: method2 over days time: method3 over years
    [...]"

    conclusion:
    we will try to find a climatological time axis first by following above hints
    and then find all variables that depend on it
    additionally, all variables with matching cell_methods will be collected
    '''
    c_time = set() # collection of climatological time axes (could be more than one per file)
    c_vars = set() # resulting list of climatological variables

    dim_vars = find_dim_vars(ncfile)

    # find all time dimension variables
    T_vars = map_dim_vars_to_types(ncfile, reverse_map=True).get('T', [])

    for vname, var in ncfile.variables.iteritems():
        is_c = False
        if vname in dim_vars:
            if hasattr(var,'climatology') and not hasattr(var,'bounds'):
                # has climatology attribute and no bounds, should be proof enough
                is_c = True
            if vname in T_vars and hasattr(var, 'units'):
                # variable is a time dimension variable and it has units
                # not a very flexible check, but this notation should not be used anyway
                units_split = var.units.split()
                if len(units_split) == 3 and units_split[1:] == ['since', '0-1-1']:
                    is_c = True
            if is_c:
                c_time.add(vname)
        else:
            # not a dimension variable, try to find out by checking cell_methods
            # this test might catch a little too many cases as the descriptions are a bit fuzzy
            if hasattr(var, 'cell_methods'):
                try:
                    cell_methods = parse_cell_methods(var.cell_methods)
                    for item in cell_methods:
                        # check if 'within' or 'over' appears in modifiers string
                        # and if 'days' or 'years' appears after it
                        # false-positives cannot be completely avoided if where ... over is used
                        splt = item[2].split()
                        if len(splt) < 2 or not (splt[-2] in ('within', 'over') and splt[-1] in ('days', 'years')):
                            break
                    else:
                        # loop did not break, so all strings found
                        is_c = True
                except:
                    pass
            if is_c:
                c_vars.add(vname)
    # now also add all variables that depend on any of the climatology time dimensions
    data_vars = set(ncfile.variables.keys()) - set(dim_vars)
    for dimname in c_time:
        for vname in data_vars:
            var = ncfile.variables[vname]
            if dimname in var.dimensions:
                c_vars.add(vname)
    return list(c_vars)

def find_compress_vars(ncfile):
    '''
    find all dimension variables that have a compress attribute and data type int
    then create a mapping between those variable names and their associated compressed dimensions
    dimensions that are not found in the file are skipped without notice

    ncfile: open input file
    returns: mapping: variable name -> compress attribute contents (compressed dimensions) as a list
    '''
    ret = {}
    for vname in find_dim_vars(ncfile):
        var = ncfile.variables[vname]
        compress = getattr(var, 'compress', None)
        if var.dtype in (np.dtype(np.int32),) and is_string_type(compress):
            dims_exist = [d for d in compress.split() if d in ncfile.dimensions]
            ret[vname] = dims_exist
    return ret

def find_auxiliary_coord_vars(ncfile, filt=lambda ncfile,c_name: True):
    '''
    find all variables which are referenced by coordinate attributes
    from CF-1.0 chapter 5:
       "The value of the coordinates attribute is a blank separated list of the names of
        auxiliary coordinate variables."
    failures are not treated here, unreadable attributes and non-existing variables
    are dropped without notice

    ncfile: open input file
    filt: filter to apply to coordinate variables to decide if they should be returned;
          default: return everything (used by find_scalar_coord_vars
    returns: mapping: variable name -> list of associated coordinate variables
    '''
    ret = {}
    for vname in ncfile.variables:
        coordinates = getattr(ncfile.variables[vname],'coordinates', None)
        if is_string_type(coordinates):
            c_vars_exist = [c_var for c_var in coordinates.split() if c_var in ncfile.variables]
            c_vars_filtered = [c_var for c_var in c_vars_exist if filt(ncfile, c_var)]
            if c_vars_filtered:
                # only add non-empty lists
                ret[vname] = c_vars_filtered
    return ret

def find_scalar_coord_vars(ncfile):
    '''
    find all scalar coordinate variables in a given file
    scalar coordinate variables are identified in the following way:
    - variable name appears in another variable's coordinates attribute
    - variable has 0 dimensions
    failuers are silently ignored
    '''
    # it is assumed that the filter is only applied to existing variable names
    filt = lambda ncfile, c_name: len(ncfile.variables[c_name].dimensions) == 0
    return find_auxiliary_coord_vars(ncfile, filt=filt)

def find_label_vars(ncfile):
    '''
    find all label variables in the given file

    label variables are identified by:
     - type is string
     - dimensions is 2
     - referenced by a coordinates attribute
     - standard_name is not "area_type"

    ncfile: open input file
    returns: list of label variables
    '''
    ret = set([])
    coord_vars = set(flatten(find_auxiliary_coord_vars(ncfile).values()))
    for vname in ncfile.variables:
        if ncfile.variables[vname].dtype.char == 'S' and \
           ncfile.variables[vname].ndim == 2 and vname in coord_vars and \
           getattr(ncfile.variables[vname],'standard_name', '') != 'area_type':
            ret.add(vname)
    return list(ret)

def find_grid_mapping_vars(ncfile):
    '''
    find all variables which are referenced by a grid_mapping attribute and exist in the file

    ncfile: open input file
    returns: mapping: variable name -> grid_mapping attribute (grid mapping variable)
    '''
    ret = {}
    for vname in ncfile.variables:
        grid_mapping = getattr(ncfile.variables[vname], 'grid_mapping', None)
        if is_string_type(grid_mapping) and grid_mapping in ncfile.variables:
            ret[vname] = grid_mapping
    return ret

def find_area_type_vars(ncfile):
    '''
    find all variables that define area_types
    identified by:
     - type is string
     - dimensions is 2
     - standard_name is "area_type"

    first mention: CF-1.4

    ncfile: open input file
    returns: list of label variables
    '''
    ret = set([])
    for vname in ncfile.variables:
        if ncfile.variables[vname].dtype.char == 'S' and \
           ncfile.variables[vname].ndim == 2 and \
           getattr(ncfile.variables[vname], 'standard_name', '') == 'area_type':
            ret.add(vname)
    return list(ret)

def get_cf_version(ncfile):
    '''
    try to automatically determine the CF version by looking at the conventions attribute of the file

    the type of the value has to be string and this string has to start with "CF-".
    everything after "CF-" will be returned as version, possibly stripped of spaces.
    if nothing matching is found, None is returned

    filename: file to check
    returns: value found in conventions attribute or None
    '''
    conventions = getattr(ncfile, 'Conventions', None)
    if is_string_type(conventions) and conventions.startswith('CF-'):
        return conventions[3:].strip() or None # replace empty strings by None
    else:
            return None

def units_known(units):
    '''
    check if given units string is accepted by udunits

    units: units string to check
    returns: units string accepted by udunits (bool)
    '''
    # non-string units can never be known
    if not is_string_type(units):
        return False
    try:
        # use 366_day calendar to avoid gregorian/julian calendar problems when parsing times
        udunits(units, calendar='366_day')
    except (TypeError, ValueError):
        return False
    else:
        return True

def units_convertible(unit1, unit2, reftimeistime=True):
    '''
    check if the two given units are convertible by UDUNITS

    unit1, unit2: unit strings to check
    reftimeistime: treat relative times and "normal" time units as convertible (default)

    returns: units can be converted into each other (bool)
    '''
    # use 366_day calendar to avoid gregorian/julian calendar problems when parsing times
    if not (units_known(unit1) and units_known(unit2)):
        raise UnknownUnitsError('"%s" and/or "%s" are not valid units according to udunits' % (unit1, unit2))
    if reftimeistime:
        # treat units as the same if both are either "normal" time or relative time
        if ( (udunits(unit1, calendar='366_day').isreftime or udunits(unit1, calendar='366_day').istime) and
              (udunits(unit2, calendar='366_day').isreftime or udunits(unit2, calendar='366_day').istime) ):
                  return True

    return bool(udunits(unit1, calendar='366_day').equivalent(udunits(unit2, calendar='366_day')))

def comp_asc(x,y):
    '''
    simple comparison operator for ascending order

    x, y: input values to compare
    returns: -1 if x < y, 0 if x == y, 1 if x > y
    '''
    if x < y:
       return -1
    elif x == y:
        return 0
    else:
        return 1

def comp_desc(x,y):
    '''
    simple comparison operator for descending order

    x, y: input values to compare
    returns: -1 if x > y, 0 if x == y, 1 if x < y
    '''
    return comp_asc(y,x)

def is_ascending(values_in):
    '''
    check if a given list of numerical values is in strictly ascending order

    values: list of values to check
    returns: list is in ascending order (bool)
    '''
    values = np.array(values_in)
    if values.ndim != 1:
        raise Exception('can only check order of 1-D arrays')
    for idx in xrange(len(values)-1):
        if not (comp_asc(values[idx], values[idx+1]) < 0):
            return False
    return True

def is_monotonic(values_in, circular=False):
    '''
    check if a given list of numerical values is in strictly monotonic order

    values: list of values to check
    circular: consider input values to belong to a circular axis (like longitude)
    return: list is in monotonic order (bool)
    '''
    # copy to numpy array for easier handling
    values = np.array(values_in)
    if values.ndim != 1:
        raise Exception('can only check order of 1-D arrays')
    elif len(values) < 2:
        # if there are less than 2 values, we consider the array is monotonic
        return True
    elif circular:
        # try to determine where the discontinuity is
        min_idx = np.where(values == values.min())[0]  # where returns a tuple with one element
        max_idx = np.where(values == values.max())[0]  # ...
        if len(min_idx) > 1 or len(max_idx) > 1:
            # if min or max occur more than once, values can never be strictly monotonic
            return False
        min_idx, max_idx = min_idx[0], max_idx[0]
        if sorted([min_idx, max_idx]) == [0, len(values)-1]:
            # no special circular handling needed
            return is_monotonic(values, circular=False)
        elif abs(min_idx-max_idx) != 1:
            # if min_idx and max_idx are not close together, values can't be monotonic
            return False
        else:
            # only one min and max value, indices are close together
            return is_monotonic(values[:min([min_idx,max_idx])], circular=False) and \
                   is_monotonic(values[max([min_idx,max_idx]):], circular=False)
    else:
        # 1-dimensional array with at least 2 values
        # determine direction from first 2 values
        comp = comp_asc if (values[0] < values[1]) else comp_desc
        # run the whole array through the comparison
        for idx in xrange(len(values) -1):
            if not (comp(values[idx], values[idx+1]) < 0):
                return False
        return True
